# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Data store implementation with backend HDF5 file."""
import array
import errno
import logging
import os
import warnings
from collections.abc import Mapping, MutableMapping
from threading import RLock

from ._dict_manager import _DictManager
from ._utility import _safe_relpath
from .errors import H5StoreAlreadyOpenError, H5StoreClosedError, InvalidKeyError

__all__ = [
    "H5Store",
    "H5Group",
    "H5StoreManager",
    "H5StoreClosedError",
    "H5StoreAlreadyOpenError",
]


def _group_is_pandas_type(group):
    return "pandas_type" in group.attrs


_is_pandas_type = None
_pandas = None


def _load_pandas():
    # Late binding to improve import performance.
    global _pandas
    global _is_pandas_type

    if _pandas is None or _is_pandas_type is None:
        try:
            import pandas

            def _is_pandas_type(value):
                return isinstance(value, pandas.core.generic.PandasObject)

        except ImportError:

            def _is_pandas_type(value):
                return False  # Must be False when pandas is not available.

        else:
            _pandas = pandas


def _requires_tables():
    try:
        import tables  # noqa
    except ImportError:
        raise ImportError(
            "Storing and loading pandas objects requires the PyTables package."
        )


logger = logging.getLogger(__name__)


def _h5set(store, grp, key, value, path=None):
    """Set a key in an h5py container.

    This method recursively converts Mappings to h5py groups and transparently
    handles None values.
    """
    import h5py
    import numpy  # h5py depends on numpy, so this is safe.

    path = path + "/" + key if path else key

    # Guard against assigning a group to itself, e.g., `h5s[key] = h5s[key]`,
    # where h5s[key] is a mapping. This is necessary, because the original
    # mapping would be deleted prior to assignment.
    if key in grp:
        if isinstance(value, H5Group):
            if grp[key] == value._group:
                return  # Groups are identical, do nothing.
        elif isinstance(value, h5py._hl.dataset.Dataset):
            if grp == value.parent:
                return  # Dataset is identical, do nothing.

        # Delete any existing data
        del grp[key]

    # Mapping-types
    if isinstance(value, Mapping):
        subgrp = grp.create_group(key)
        for k, v in value.items():
            _h5set(store, subgrp, k, v, path)

    # Regular built-in types:
    elif value is None:
        grp.create_dataset(key, data=None, shape=None, dtype="f")
    elif isinstance(value, (int, float, str, bool, array.array)):
        grp[key] = value
    elif isinstance(value, bytes):
        grp[key] = numpy.bytes_(value)

    # NumPy types
    elif type(value).__module__ == numpy.__name__:
        grp[key] = value

    # h5py native types
    elif isinstance(value, h5py._hl.dataset.Dataset):
        grp[key] = value  # Creates hard-link!

    # Other types
    else:
        _load_pandas()  # might be a pandas type
        if _is_pandas_type(value):
            _requires_tables()
            store.close()
            with _pandas.HDFStore(store.filename, mode="a") as store_:
                store_[path] = value
            store.open()
        else:
            grp[key] = value
            warnings.warn(
                "Storage for object of type '{}' appears to have succeeded, but this "
                "type is not officially supported!".format(type(value))
            )


def _h5get(store, grp, key, path=None):
    """Retrieve the underlying data for a key from its h5py container."""
    import h5py

    path = path + "/" + key if path else key
    result = grp[key]

    if _group_is_pandas_type(result):
        _load_pandas()
        _requires_tables()
        grp.file.flush()
        # The store must be closed for pandas to open it safely, but first we
        # copy the filename since it is not accessible after closing the file.
        # The pandas data is returned by copy, so the HDFStore can be closed.
        # Then we re-open the store.
        filename = grp.file.filename
        store.close()
        with _pandas.HDFStore(filename, mode="r") as store_:
            data = store_[path]
        store.open()
        return data
    try:
        shape = result.shape
        if shape is None:
            return None
        elif shape:
            return result
        elif (
            h5py.version.version_tuple.major >= 3
            and h5py.check_dtype(vlen=result.dtype) is str
        ):
            # h5py >=3.0.0 returns strings as bytes. This returns str for
            # consistency with past behavior in signac.
            return result.asstr()[()]
        else:
            return result[()]
    except AttributeError:
        if isinstance(result, MutableMapping):
            return H5Group(store, path)
        else:
            return result


class _ensure_open:
    __slots__ = ["file", "open", "kwargs"]

    def __init__(self, file, **kwargs):
        self.file = file
        self.open = False
        self.kwargs = kwargs

    def __enter__(self):
        if self.file._file is None:
            self.file._open(**self.kwargs)
            self.open = True

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self.open:
            self.file.close()
            self.open = False


class H5Group(MutableMapping):
    """An abstraction layer over h5py's Group objects, to manage and return data."""

    __slots__ = ["_store", "_path"]

    def __init__(self, store, path):
        self._store = store
        self._path = path

    def __repr__(self):
        return "{}(store={}, path={})".format(
            type(self).__name__, repr(self._store), repr(self._path)
        )

    @property
    def _group(self):
        return self._store.file[self._path]

    def __getitem__(self, key):
        with _ensure_open(self._store):
            return _h5get(self._store, self._group, key, self._path)

    def __setitem__(self, key, value):
        with _ensure_open(self._store):
            _h5set(
                self._store,
                self._group,
                self._store._validate_key(key),
                value,
                self._path,
            )
            return value

    def __delitem__(self, key):
        with _ensure_open(self._store):
            del self._group[key]

    def __getattr__(self, name):
        with _ensure_open(self._store):
            if name in self._group.keys():
                return self.__getitem__(name)
            else:
                return getattr(self._group, name)

    def __setattr__(self, key, value):
        if key.startswith("__") or key in self.__slots__:
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def setdefault(self, key, value):
        """Set a value for a key if that key is not already set."""
        super().setdefault(key, value)
        return self.__getitem__(key)

    def __iter__(self):
        with _ensure_open(self._store):
            yield from self._group.keys()

    def __len__(self):
        with _ensure_open(self._store):
            return len(self._group)

    def __eq__(self, other):
        with _ensure_open(self._store):
            if isinstance(self, Mapping) and isinstance(other, Mapping):
                return super().__eq__(other)
            elif type(other) == type(self):
                return self._group == other._group
            else:
                return super().__eq__(other)


class H5Store(MutableMapping):
    r"""An HDF5-backed container for storing array-like and dictionary-like data.

    The H5Store is a :class:`~collections.abc.MutableMapping` and therefore
    behaves similar to a :class:`dict`, but all data is stored persistently in
    the associated HDF5 file on disk.

    Supported types include:

      * built-in types (int, float, str, bool, NoneType, array)
      * numpy arrays
      * pandas data frames (requires pandas and pytables)
      * mappings with values that are supported types

    Values can be accessed as attributes (``h5s.foo``) or via key index
    (``h5s['foo']``).

    Examples
    --------
    >>> from signac import H5Store
    >>> with H5Store('file.h5') as h5s:
    ...     h5s['foo'] = 'bar'
    ...     assert 'foo' in h5s
    ...     assert h5s.foo == 'bar'
    ...     assert h5s['foo'] == 'bar'

    The H5Store can be used as a context manager to ensure that the underlying
    file is opened, however most built-in types (excluding arrays) can be read
    and stored without the need to *explicitly* open the file. **To
    access arrays (reading or writing), the file must always be opened!**

    To open a file in read-only mode, use the :py:meth:`.open` method with ``mode='r'``:

    >>> with H5Store('file.h5').open(mode='r') as h5s:
    ...     pass

    Parameters
    ----------
    filename : str
        The filename of the underlying HDF5 file.
    \*\*kwargs
        Additional keyword arguments to be forwarded to the
        :py:class:`h5py.File` constructor. See the :py:class:`h5py.File`
        documentation for more information.

    """
    __slots__ = ["_filename", "_file", "_kwargs"]

    _thread_lock = RLock()

    def __init__(self, filename, **kwargs):
        if not (isinstance(filename, str) and len(filename) > 0):
            raise ValueError("H5Store filename must be a non-empty string.")
        self._filename = os.path.abspath(filename)
        self._file = None
        self._kwargs = kwargs

    @property
    def filename(self):
        """Return the H5Store filename."""
        return self._filename

    def __repr__(self):
        return f"{type(self).__name__}(filename={repr(_safe_relpath(self.filename))})"

    def __str__(self):
        return "{}(filename={})".format(
            type(self).__name__, repr(os.path.basename(self.filename))
        )

    def __del__(self):
        self.close()

    def __enter__(self):
        try:
            self.open()
        except H5StoreAlreadyOpenError:
            pass
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def _open(self, **kwargs):
        if self._file is not None:
            raise H5StoreAlreadyOpenError(self)
        import h5py

        # We use the default file parameters, which can optionally be overridden
        # by additional keyword arguments (kwargs). This option is intentionally
        # not exposed to the public API.
        parameters = dict(self._kwargs)
        parameters.update(kwargs)
        if parameters.get("mode", None) is None:
            parameters["mode"] = "a"

        self._thread_lock.acquire()
        try:
            self._file = h5py.File(self.filename, **parameters)
        except:  # noqa We need to release under **all** circumstances upon error!
            self._thread_lock.release()
            raise
        return self

    def open(self, mode=None):
        """Open the underlying HDF5 file.

        Parameters
        ----------
        mode : str
            The file open mode to use. Defaults to 'a' (append).

        Returns
        -------
        H5Store
            This H5Store instance.
        """
        if mode is None:
            mode = self._kwargs.get("mode", "a")
        return self._open(mode=mode)

    def close(self):
        """Close the underlying HDF5 file."""
        locked = True
        try:
            self._file.close()
            self._file = None
        except AttributeError:
            locked = False
        finally:
            if locked:
                try:
                    self._thread_lock.release()
                except RuntimeError as error:
                    if "cannot release un-acquired lock" not in str(error):
                        raise

    @property
    def file(self):
        """Access the underlying instance of :py:class:`h5py.File`.

        This property exposes the underlying :py:class:`h5py.File` object, enabling
        use of functions such as :py:meth:`h5py.Group.create_dataset` or
        :py:meth:`h5py.Group.require_dataset`.

        .. note::

            The store must be open to access this property!

        Returns
        -------
        h5py.File
            The :py:class:`h5py.File` object that this store is operating on.

        Raises
        ------
        H5StoreClosedError
            If the store is closed.
        """
        if self._file is None:
            raise H5StoreClosedError(self.filename)
        else:
            return self._file

    @property
    def mode(self):
        """Return the default opening mode of this H5Store."""
        return self.file.mode

    def flush(self):
        """Flush the underlying HDF5 file."""
        if self._file is None:
            raise H5StoreClosedError(self.filename)
        else:
            self._file.flush()

    def __getitem__(self, key):
        key = key if key.startswith("/") else "/" + key
        with _ensure_open(self):
            return _h5get(self, self._file, key)

    @staticmethod
    def _validate_key(key):
        """Emit a warning or raise an exception if key is invalid. Returns key."""
        if "." in key:
            raise InvalidKeyError("Keys for the H5Store may not contain dots ('.').")
        return key

    def __setitem__(self, key, value):
        with _ensure_open(self):
            _h5set(self, self._file, self._validate_key(key), value)
            return value

    def __delitem__(self, key):
        with _ensure_open(self):
            del self._file[key]

    def __getattr__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name.startswith("__") or name in self.__slots__:
                raise
            try:
                return self.__getitem__(name)
            except KeyError as e:
                raise AttributeError(e)

    def __setattr__(self, key, value):
        if key.startswith("__") or key in self.__slots__:
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def __delattr__(self, key):
        if key.startswith("__") or key in self.__slots__:
            super().__delattr__(key)
        else:
            self.__delitem__(key)

    def setdefault(self, key, value):
        """Set a value for a key if that key is not already set."""
        super().setdefault(key, value)
        return self.__getitem__(key)

    def __iter__(self):
        with _ensure_open(self):
            yield from self._file.keys()

    def __len__(self):
        try:
            with _ensure_open(self, mode="r"):
                return len(self._file)
        except OSError as error:
            if f"errno = {errno.ENOENT}" in str(error):
                return 0  # file does not exist
            else:
                raise

    def __contains__(self, key):
        try:
            with _ensure_open(self, mode="r"):
                return key in self._file
        except OSError as error:
            if f"errno = {errno.ENOENT}" in str(error):
                return False  # file does not exist
            else:
                raise

    def clear(self):
        """Remove all data from this store.

        .. danger::
            All data will be removed, this action cannot be reversed!
        """
        with _ensure_open(self):
            self._file.clear()


class H5StoreManager(_DictManager):
    """Helper class to manage multiple instances of :class:`~.H5Store` within a directory.

    Parameters
    ----------
    prefix : str
        The directory prefix shared by all files managed by this class.

    Examples
    --------
    Assuming that the ``stores/`` directory exists:

    >>> stores = H5StoreManager('stores/')
    >>> stores.data
    <H5Store(filename=stores/data.h5)>
    >>> stores.data.foo = True
    >>> dict(stores.data)
    {'foo': True}
    """

    cls = H5Store  # type: ignore
    suffix = ".h5"  # type: ignore

    @staticmethod
    def _validate_key(key):
        """Emit a warning or raise an exception if key is invalid. Returns key."""
        if "." in key:
            raise InvalidKeyError(
                "Keys for the H5StoreManager may not contain dots ('.')."
            )
        return key



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Project Schema."""

from collections.abc import Mapping
from numbers import Number
from pprint import pformat

from ._search_indexer import _DictPlaceholder
from ._utility import _nested_dicts_to_dotted_keys


class _Vividict(dict):
    """A dict that returns an empty _Vividict for keys that are missing.

    Useful for automatically nesting keys with ``vividict['a']['b']['c'] = 'd'``.

    """

    def __missing__(self, key):
        value = self[key] = type(self)()
        return value


def _strip_prefix(key):
    return key[len("sp.") :]


def _build_job_statepoint_index(exclude_const, index):
    """Build index for job state points.

    Parameters
    ----------
    exclude_const : bool
        Excludes state point keys whose values are constant across the index.
    index : _SearchIndexer
        A _SearchIndexer mapping from job ids to job state points.

    Yields
    ------
    statepoint_key : str
        State point key.
    statepoint_values : dict
        Dictionary mapping from a state point value to the set of job ids with
        that state point value.

    """
    indexes = {}
    for _id in index.find():
        doc = index[_id]
        for key, _ in _nested_dicts_to_dotted_keys(doc):
            if key.split(".")[0] == "sp":
                indexes[key] = index.build_index(key)

    for key in sorted(indexes, key=lambda key: (len(indexes[key]), key)):
        if (
            exclude_const
            and len(indexes[key]) == 1
            and len(indexes[key][next(indexes[key].keys())]) == len(index)
        ):
            continue
        statepoint_key = _strip_prefix(key)
        # Remove _DictPlaceholder keys from the index
        statepoint_values = indexes[key]
        statepoint_values.pop(_DictPlaceholder, None)
        yield statepoint_key, statepoint_values


class ProjectSchema(Mapping):
    """A description of a project's state point schema.

    Parameters
    ----------
    schema : dict
        Project schema.

    """

    def __init__(self, schema=None):
        if schema is None:
            schema = {}
        self._schema = schema

    def format(self, depth=None, precision=None, max_num_range=None):
        """Format the schema for printing.

        Parameters
        ----------
        depth : int
            A non-zero value will return a nested formatting up to the specified depth,
            defaults to 0.
        precision : int
            Round numerical values to the given precision, defaults to unlimited precision.
        max_num_range : int
            The maximum number of entries shown for a value range, defaults to 5.

        Returns
        -------
        str
            A formatted representation of the project schema.

        """
        if depth is None:
            depth = 0
        if max_num_range is None:
            max_num_range = 5

        def _fmt_value(x):
            """Convert a value to a string, rounded to a given precision.

            Parameters
            ----------
            x :
                Value to convert to string.

            Returns
            -------
            str
                Formatted value.

            """
            if precision is not None and isinstance(x, Number):
                return str(round(x, precision))
            else:
                return str(x)

        def _fmt_range(type_, values):
            """Convert sequence of values into a comma-separated string.

            Inserts an ellipsis (...) if the number of values exceeds ``max_num_range``.

            Parameters
            ----------
            type_ : type
                Type of values.
            values :
                An iterable of values.

            Returns
            -------
            str
                Comma-separated string of the input values.

            """
            try:
                sorted_values = sorted(values)
            except TypeError:
                sorted_values = sorted(values, key=repr)
            if len(values) <= max_num_range:
                values_string = ", ".join(_fmt_value(value) for value in sorted_values)
            else:
                values_string = ", ".join(
                    _fmt_value(value) for value in sorted_values[: max_num_range - 2]
                )
                values_string += ", ..., "
                values_string += ", ".join(_fmt_value(v) for v in sorted_values[-2:])
            return "{type_name}([{values_string}], {length})".format(
                type_name=type_.__name__,
                values_string=values_string,
                length=len(values),
            )

        def _fmt_values(values):
            """Convert values into a single string.

            Parameters
            ----------
            values :
                An iterable of values.

            Returns
            -------
            str
                Comma-separated string of the input values.

            """
            return ", ".join(_fmt_range(*value) for value in values.items())

        if depth > 0:
            schema_dict = _Vividict()
            for key, values in self._schema.items():
                keys = key.split(".")
                if len(keys) > 1:
                    x = schema_dict[keys[0]]
                    for k in keys[1:-1]:
                        x = x[k]
                    x[keys[-1]] = _fmt_values(values)
                else:
                    schema_dict[keys[0]] = _fmt_values(values)
            return pformat(schema_dict, depth=depth)
        else:
            ret = ["{"]
            for key in sorted(self._schema):
                values = self._schema[key]
                if values:
                    ret.append(f" '{key}': '{_fmt_values(values)}',")
            ret.append("}")
            return "\n".join(ret)

    def __len__(self):
        return len(self._schema)

    def __str__(self):
        return self.format()

    def __repr__(self):
        return f"{type(self).__name__}(<len={len(self)}>)"

    def _repr_html_(self):
        import html

        output = "<strong>" + html.escape(repr(self)) + "</strong>"
        output += "<pre>" + str(self) + "</pre>"
        return output

    def __getitem__(self, key_or_keys):
        return self._schema[key_or_keys]

    def __iter__(self):
        return iter(self._schema)

    def difference(self, other, ignore_values=False):
        """Determine the difference between this and another project schema.

        Parameters
        ----------
        ignore_values : bool, optional
            Ignore if the value (range) of a specific keys differ,
            only return missing keys (Default value = False).
        other : :class:`~ProjectSchema`
            Other project schema.

        Returns
        -------
        set
            A set of key tuples that are either missing or different in the other schema.

        """
        ret = set(self.keys()).difference(other.keys())
        if not ignore_values:
            ret.update(
                {
                    key
                    for key, value in self.items()
                    if key in other and other[key] != value
                }
            )
        return ret



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os
import platform
import random
import string
import subprocess
import sys
from array import array
from collections.abc import Mapping
from contextlib import closing, contextmanager
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from platform import python_implementation
from tempfile import TemporaryDirectory
from time import time

import pytest

from signac.errors import InvalidKeyError
from signac.h5store import H5Store, H5StoreAlreadyOpenError, H5StoreClosedError

PYPY = "PyPy" in platform.python_implementation()

try:
    import h5py  # noqa

    H5PY = True
except ImportError:
    H5PY = False

try:
    import pandas  # noqa
    import tables  # noqa

    PANDAS_AND_TABLES = True
except ImportError:
    PANDAS_AND_TABLES = False

try:
    import numpy  # noqa

    NUMPY = True
except ImportError:
    NUMPY = False

FN_STORE = "signac_test_h5store.h5"


WINDOWS = sys.platform == "win32"


@pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
@pytest.mark.skipif(PYPY, reason="h5py not reliable on PyPy platform")
class TestH5StoreBase:
    @pytest.fixture(autouse=True)
    def setUp_base_h5Store(self, request):
        self._tmp_dir = TemporaryDirectory(prefix="signac_test_h5store_")
        request.addfinalizer(self._tmp_dir.cleanup)
        self._fn_store = os.path.join(self._tmp_dir.name, FN_STORE)
        self._fn_store_other = os.path.join(self._tmp_dir.name, "other_" + FN_STORE)

    def get_h5store(self, **kwargs):
        return H5Store(filename=self._fn_store, **kwargs)

    @contextmanager
    def open_h5store(self, **kwargs):
        with self.get_h5store(**kwargs) as h5s:
            yield h5s

    def get_other_h5store(self, **kwargs):
        return H5Store(filename=self._fn_store_other, **kwargs)

    @contextmanager
    def open_other_h5store(self, **kwargs):
        with self.get_other_h5store(**kwargs) as h5s:
            yield h5s

    def get_testdata(self, size=None):
        if size is None:
            size = 1024
        return "".join(random.choice(string.ascii_lowercase) for i in range(size))

    def assertEqual(self, a, b):
        if hasattr(a, "shape"):
            if not NUMPY:
                raise pytest.skip("This test requires the numpy package.")
            numpy.testing.assert_array_equal(a, b)
        else:
            assert a == b


class TestH5StoreOpen(TestH5StoreBase):
    def test_open(self):
        h5s = self.get_h5store()
        h5s.open()
        h5s.close()

    def test_open_read_only(self):
        with self.open_h5store() as h5s:
            h5s["foo"] = "bar"

        with self.open_h5store(mode="r") as h5s:
            assert h5s.mode == "r"
            assert "foo" in h5s
            self.assertEqual(h5s["foo"], "bar")

    def test_open_write_only(self):
        with self.open_h5store(mode="w") as h5s:
            assert h5s.mode == "r+"
            h5s["foo"] = "bar"
            assert "foo" in h5s
            self.assertEqual(h5s["foo"], "bar")

    def test_open_write_and_read_only(self):
        with self.open_h5store(mode="w") as h5s_w:
            with self.open_h5store(mode="r") as h5s_r:
                assert h5s_w.mode == "r+"
                assert h5s_r.mode == "r+"
                assert "foo" not in h5s_r
                assert "foo" not in h5s_w

                h5s_w["foo"] = "bar"
                assert "foo" in h5s_r
                self.assertEqual(h5s_r["foo"], "bar")
                assert "foo" in h5s_w
                self.assertEqual(h5s_r["foo"], "bar")


class TestH5Store(TestH5StoreBase):
    valid_types = {
        "int": 123,
        "float": 123.456,
        "string": "lorem ipsum including unicõdé 🎨",
        "none": None,
        "float_array": array("f", [-1.5, 0, 1.5]),
        "double_array": array("d", [-1.5, 0, 1.5]),
        "int_array": array("i", [-1, 0, 1]),
        "uint_array": array("I", [0, 1, 2]),
        "dict": {"a": 1, "b": None, "c": "test"},
        "bytes": b"\x73\x69\x67\x6E\x61\x63\x00\xF0\x9F\x8E\xA8",
    }

    if NUMPY:
        valid_types.update(
            {
                "numpy_int_array": numpy.array([-1, 0, 1], dtype=numpy.int_),
                "numpy_float_array": numpy.array([-1.5, 0, 1.5], dtype=numpy.float_),
                "numpy_complex_array": numpy.array(
                    [-1.5 + 3.14j, 0, 1.5 - 5.67j], dtype=numpy.complex_
                ),
                # Note that NumPy's string type is handled kind of like a char
                # array or bytes, not like Python str. numpy.unicode_ uses a
                # four byte encoding and is explicitly not supported by h5py.
                "numpy_string_array": numpy.array(
                    [
                        b"abcde",
                        b"\x73\x69\x67\x6E\x61\x63\x00\xF0\x9F\x8E\xA8",
                    ],
                    dtype=numpy.string_,
                ),
                "numpy_void_array": numpy.array(
                    [
                        b"abcdefghijk",
                        b"\x73\x69\x67\x6E\x61\x63\x00\xF0\x9F\x8E\xA8",
                    ],
                    dtype="V",
                ),
            }
        )

    def test_init(self):
        self.get_h5store()

    def test_invalid_filenames(self):
        with pytest.raises(ValueError):
            H5Store(None)
        with pytest.raises(ValueError):
            H5Store("")
        with pytest.raises(ValueError):
            H5Store(123)

    def test_set_get(self):
        with self.open_h5store() as h5s:
            key = "setget"
            d = self.get_testdata()
            h5s.clear()
            assert not bool(h5s)
            assert len(h5s) == 0
            assert key not in h5s
            with pytest.raises(KeyError):
                h5s[key]
            d_ = h5s[key] = d
            self.assertEqual(d_, d)
            assert bool(h5s)
            assert len(h5s) == 1
            assert key in h5s
            self.assertEqual(h5s[key], d)
            self.assertEqual(h5s.get(key), d)
            assert h5s.get("nonexistent", "default") == "default"

    def test_set_get_explicit_nested(self):
        with self.open_h5store() as h5s:
            key = "setgetexplicitnested"
            d = self.get_testdata()
            assert "a" not in h5s
            ret = h5s.setdefault("a", dict())
            assert "a" in h5s
            self.assertEqual(ret, h5s["a"])
            assert hasattr(ret, "_store")  # is an H5Group object
            child1 = h5s["a"]
            child2 = h5s["a"]
            self.assertEqual(child1, child2)
            assert type(child1) == type(child2)
            assert not child1
            assert not child2
            child1[key] = d
            assert child1
            assert child2
            assert key in child1
            assert key in child2
            self.assertEqual(child1, child2)
            self.assertEqual(child1[key], d)
            self.assertEqual(child2[key], d)

    def test_setdefault_group(self):
        with self.open_h5store() as h5s:
            group_key = "setdefault_group"
            key = "setdefault"
            d = self.get_testdata()
            assert group_key not in h5s
            h5s[group_key] = {}
            assert group_key in h5s
            assert len(h5s[group_key]) == 0
            assert key not in h5s[group_key]
            h5s[group_key].setdefault(key, d)
            assert len(h5s[group_key]) == 1
            assert key in h5s[group_key]
            self.assertEqual(h5s[group_key][key], d)
            d2 = self.get_testdata()
            h5s[group_key].setdefault(key, d2)
            self.assertEqual(h5s[group_key][key], d)

    def test_repr(self):
        with self.open_h5store() as h5s:
            key = "test_repr"
            assert repr(h5s) == repr(eval(repr(h5s)))
            h5s[key] = self.get_testdata()
        assert repr(h5s) == repr(eval(repr(h5s)))

    def test_str(self):
        with self.open_h5store() as h5s:
            key = "test_repr"
            h5s[key] = self.get_testdata()
            str(h5s)  # open
        str(h5s)  # closed

    def test_len(self):
        h5s = self.get_h5store()
        assert len(h5s) == 0
        h5s["test_len"] = True
        assert len(h5s) == 1

    def test_contains(self):
        h5s = self.get_h5store()
        assert "test_contains" not in h5s
        h5s["test_contains"] = True
        assert "test_contains" in h5s

    def test_copy_value(self):
        with self.open_h5store() as h5s:
            key = "copy_value"
            key2 = "copy_value2"
            d = self.get_testdata()
            h5s[key] = d
            assert key in h5s
            self.assertEqual(h5s[key], d)
            assert key2 not in h5s
            h5s[key2] = h5s[key]
            assert key in h5s
            self.assertEqual(h5s[key], d)
            assert key2 in h5s
            self.assertEqual(h5s[key2], d)

    def test_iter(self):
        with self.open_h5store() as h5s:
            key1 = "iter1"
            key2 = "iter2"
            d1 = self.get_testdata()
            d2 = self.get_testdata()
            d = {key1: d1, key2: d2}
            h5s.update(d)
            assert key1 in h5s
            assert key2 in h5s
            for i, key in enumerate(h5s):
                assert key in d
                self.assertEqual(d[key], h5s[key])
            assert i == 1

    def test_delete(self):
        with self.open_h5store() as h5s:
            key = "delete"
            d = self.get_testdata()
            h5s[key] = d
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d)
            del h5s[key]
            assert len(h5s) == 0
            with pytest.raises(KeyError):
                h5s[key]

    def test_delete_attr(self):
        with self.open_h5store() as h5s:
            key = "delete"
            d = self.get_testdata()
            h5s[key] = d
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d)
            delattr(h5s, key)
            assert len(h5s) == 0
            with pytest.raises(KeyError):
                h5s[key]

    def test_delete_nested(self):
        with self.open_h5store() as h5s:
            key1 = "group_key"
            key2 = "delete"
            d = self.get_testdata()
            h5s[key1] = {key2: d}
            assert len(h5s) == 1
            assert len(h5s[key1]) == 1
            self.assertEqual(h5s[key1][key2], d)
            del h5s[key1][key2]
            assert len(h5s) == 1
            assert len(h5s[key1]) == 0
            self.assertEqual(h5s[key1], {})
            with pytest.raises(KeyError):
                h5s[key1][key2]

    def test_update(self):
        with self.open_h5store() as h5s:
            key = "update"
            d = {key: self.get_testdata()}
            h5s.update(d)
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d[key])

    def test_clear(self):
        with self.open_h5store() as h5s:
            h5s.clear()
            key = "clear"
            d = self.get_testdata()
            h5s[key] = d
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d)
            h5s.clear()
            assert len(h5s) == 0

    def test_reopen(self):
        with self.open_h5store() as h5s:
            key = "reopen"
            d = self.get_testdata()
            h5s[key] = d
        with self.open_h5store() as h5s:
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d)

    def test_open_twice(self):
        h5s = self.get_h5store()
        h5s.open()
        try:
            with pytest.raises(H5StoreAlreadyOpenError):
                h5s.open()
        finally:
            h5s.close()

    def test_open_reentry(self):
        with self.open_h5store() as h5s:
            with h5s:
                pass

    def test_reopen_explicit_open_close(self):
        h5s = self.get_h5store().open()
        key = "reopen"
        d = self.get_testdata()
        h5s[key] = d
        h5s.close()
        h5s.open()
        assert len(h5s) == 1
        self.assertEqual(h5s[key], d)
        h5s.close()

    def test_write_valid_types(self):
        with self.open_h5store() as h5s:
            for k, v in self.valid_types.items():
                h5s[k] = v
                self.assertEqual(h5s[k], v)

    def test_assign_valid_types_within_identical_file(self):
        with self.open_h5store() as h5s:
            for k, v in self.valid_types.items():
                h5s[k] = v
                self.assertEqual(h5s[k], v)
                h5s[k] = h5s[k]
                self.assertEqual(h5s[k], v)

                k_other = k + "-other"
                h5s[k_other] = h5s[k]
                self.assertEqual(h5s[k], v)
                self.assertEqual(h5s[k_other], v)
                self.assertEqual(h5s[k], h5s[k_other])

    def test_assign_valid_types_within_same_file(self):
        with self.open_h5store() as h5s:
            with self.open_h5store() as same_h5s:
                for k, v in self.valid_types.items():
                    # First, store v under k
                    h5s[k] = v
                    self.assertEqual(h5s[k], v)

                    # Assign the same value under the same key.
                    try:
                        same_h5s[k] = h5s[k]
                    except H5StoreClosedError:
                        pass
                    self.assertEqual(h5s[k], v)
                    self.assertEqual(same_h5s[k], v)
                    self.assertEqual(h5s[k], same_h5s[k])

                    # Assign the same value, under a different key.
                    other_key = k + "-other"
                    try:
                        same_h5s[other_key] = h5s[k]
                    except H5StoreClosedError:
                        pass
                    self.assertEqual(h5s[other_key], v)
                    self.assertEqual(same_h5s[other_key], v)
                    self.assertEqual(h5s[k], same_h5s[other_key])

                    # Deleting the value assigned to the alternative key should have
                    # no effect on the value stored under the original key, regardless
                    # whether it was copied by reference (hard-link) or copied by
                    # value (true copy).
                    del same_h5s[other_key]
                    self.assertEqual(h5s[k], v)
                    self.assertEqual(same_h5s[k], v)
                    self.assertEqual(same_h5s[k], h5s[k])

    def test_assign_valid_types_between_files(self):
        with self.open_h5store() as h5s:
            with self.open_other_h5store() as other_h5s:
                for k, v in self.valid_types.items():
                    h5s[k] = v
                    self.assertEqual(h5s[k], v)
                    try:
                        other_h5s[k] = h5s[k]
                    except (OSError, RuntimeError) as error:
                        # Type of error may depend on platform or software versions
                        assert (
                            str(error)
                            == "Unable to create link (interfile hard links are not allowed)"
                        )
                        assert isinstance(v, (array, numpy.ndarray))
                        other_h5s[k] = h5s[k][()]
                    self.assertEqual(h5s[k], v)
                    self.assertEqual(other_h5s[k], v)
                    self.assertEqual(other_h5s[k], h5s[k])

    def test_write_invalid_type(self):
        class Foo:
            pass

        with self.open_h5store() as h5s:
            key = "write_invalid_type"
            d = self.get_testdata()
            h5s[key] = d
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d)
            d2 = Foo()
            with pytest.raises(TypeError):
                h5s[key + "2"] = d2
            assert len(h5s) == 1
            self.assertEqual(h5s[key], d)

    def test_keys_with_dots(self):
        with pytest.raises(InvalidKeyError):
            with self.open_h5store() as h5s:
                key = "a.b"
                d = self.get_testdata()
                h5s[key] = d
                self.assertEqual(h5s[key], d)

    def test_keys_with_slashes(self):
        # HDF5 uses slashes for nested keys internally
        with self.open_h5store() as h5s:
            key = "a/b"
            d = self.get_testdata()
            h5s[key] = d
            self.assertEqual(h5s[key], d)
            self.assertEqual(h5s["a"]["b"], d)

    def test_value_none(self):
        with self.get_h5store() as h5s:
            key = "a"
            d = None
            h5s[key] = d
            self.assertEqual(h5s[key], d)

    def test_set_get_attr_sync(self):
        with self.get_h5store() as h5s:
            assert len(h5s) == 0
            assert "a" not in h5s
            with pytest.raises(AttributeError):
                h5s.a
            a = 0
            h5s.a = a
            assert len(h5s) == 1
            assert "a" in h5s
            self.assertEqual(h5s.a, a)
            self.assertEqual(h5s["a"], a)
            a = 1
            h5s.a = a
            assert len(h5s) == 1
            assert "a" in h5s
            self.assertEqual(h5s.a, a)
            self.assertEqual(h5s["a"], a)

            def check_nested(a, b):
                assert len(h5s) == 1
                assert len(h5s.a) == 1
                assert "a" in h5s
                assert "b" in h5s.a
                self.assertEqual(h5s.a, a)
                self.assertEqual(h5s["a"]["b"], b)
                self.assertEqual(h5s.a.b, b)
                self.assertEqual(h5s["a"], a)

            h5s.a = {"b": 0}
            check_nested({"b": 0}, 0)
            h5s.a.b = 1
            check_nested({"b": 1}, 1)
            h5s["a"] = {"b": 2}
            check_nested({"b": 2}, 2)
            h5s["a"]["b"] = 3
            check_nested({"b": 3}, 3)

    def test_modify_nested(self):
        with self.get_h5store() as h5s:
            h5s.a = dict(b=True)
            a = h5s.a
            a["b"] = False
            assert not h5s.a["b"]

    def test_invalid_attr(self):
        h5s = self.get_h5store()
        with pytest.raises(AttributeError):
            h5s.a
        with pytest.raises(AttributeError):
            h5s._a
        with pytest.raises(AttributeError):
            h5s.__a__

    def test_attr_reference_modification(self):
        with self.get_h5store() as h5s:
            assert len(h5s) == 0
            assert "a" not in h5s
            with pytest.raises(AttributeError):
                h5s.a
            pairs = [(0, 1), (0.0, 1.0), ("0", "1"), (False, True)]
            dict_pairs = [(dict(c=a), dict(c=b)) for a, b in pairs]
            for A, B in chain(pairs, dict_pairs):
                h5s.a = A
                a = h5s.a
                self.assertEqual(a, A)
                self.assertEqual(h5s.a, A)
                a = B
                self.assertEqual(a, B)
                self.assertEqual(h5s.a, A)
                a = h5s["a"]
                self.assertEqual(a, A)
                self.assertEqual(h5s.a, A)
                a = B
                self.assertEqual(a, B)
                self.assertEqual(h5s.a, A)

                # with nested values
                h5s["a"] = dict(b=A)
                self.assertEqual(h5s.a.b, A)
                b = h5s.a.b
                self.assertEqual(b, A)
                self.assertEqual(h5s.a.b, A)
                b = B
                self.assertEqual(b, B)
                self.assertEqual(h5s.a.b, A)
                b = h5s["a"]["b"]
                self.assertEqual(b, A)
                self.assertEqual(h5s.a.b, A)
                b = B
                self.assertEqual(b, B)
                self.assertEqual(h5s.a.b, A)
                b = h5s["a"].b
                self.assertEqual(b, A)
                self.assertEqual(h5s.a.b, A)
                b = B
                self.assertEqual(b, B)
                self.assertEqual(h5s.a.b, A)


class TestH5StoreNestedData(TestH5Store):
    def get_testdata(self, size=None):
        return dict(a=super().get_testdata(size))

    def test_repr(self):
        from signac.h5store import H5Group, H5Store  # noqa:F401

        with self.open_h5store() as h5s:
            key = "test_repr"
            assert repr(h5s) == repr(eval(repr(h5s)))
            h5s[key] = self.get_testdata()
            assert repr(h5s[key]) == repr(eval(repr(h5s[key])))
        assert repr(h5s) == repr(eval(repr(h5s)))


class TestH5StoreBytesData(TestH5Store):
    def get_testdata(self, size=None):
        return super().get_testdata(size=size).encode()


class TestH5StoreClosed(TestH5Store):
    valid_types = {
        "int": 123,
        "float": 123.456,
        "string": "foobar",
        "none": None,
        "dict": {"a": 1, "b": None, "c": "test"},
    }

    @contextmanager
    def open_h5store(self, **kwargs):
        yield self.get_h5store(**kwargs)


class TestH5StoreNestedDataClosed(TestH5StoreNestedData, TestH5StoreClosed):
    pass


@pytest.mark.skipif(not PANDAS_AND_TABLES, reason="requires pandas and pytables")
@pytest.mark.skipif(not NUMPY, reason="requires numpy package")
class TestH5StorePandasData(TestH5Store):
    def get_testdata(self, size=None):
        if size is None:
            size = 1024
        return pandas.DataFrame(
            numpy.random.rand(8, size),
            index=[string.ascii_letters[i] for i in range(8)],
        )

    def assertEqual(self, a, b):
        if isinstance(a, Mapping):
            assert isinstance(b, Mapping)
            super().assertEqual(a.keys(), b.keys())
            for key in a:
                self.assertEqual(a[key], b[key])
        else:
            try:
                return (a == b).all()
            except (AttributeError, ValueError):
                return super().assertEqual(a, b)
            else:
                assert isinstance(a, pandas.DataFrame)


@pytest.mark.skipif(not PANDAS_AND_TABLES, reason="requires pandas and pytables")
@pytest.mark.skipif(not NUMPY, reason="requires numpy package")
class TestH5StoreNestedPandasData(TestH5StorePandasData):
    def get_testdata(self, size=None):
        if size is None:
            size = 1024
        return dict(
            df=pandas.DataFrame(
                numpy.random.rand(8, size),
                index=[string.ascii_letters[i] for i in range(8)],
            )
        )


class TestH5StoreMultiThreading(TestH5StoreBase):
    @pytest.mark.skip(
        reason="This test fails randomly on CI. "
        "See https://github.com/glotzerlab/signac/pull/307"
    )
    def test_multithreading(self):
        def set_x(x):
            self.get_h5store()["x"] = x

        with closing(ThreadPool(2)) as pool:
            pool.map(set_x, range(100))
        pool.join()

        assert self.get_h5store()["x"] in set(range(100))

    @pytest.mark.skip(
        reason="This test fails randomly on CI. "
        "See https://github.com/glotzerlab/signac/pull/307"
    )
    def test_multithreading_with_error(self):
        def set_x(x):
            self.get_h5store()["x"] = x
            if x == 50:
                raise RuntimeError()

        with pytest.raises(RuntimeError):
            with closing(ThreadPool(2)) as pool:
                pool.map(set_x, range(100))
        pool.join()

        assert self.get_h5store()["x"] in set(range(100))


def _read_from_h5store(filename, **kwargs):
    with H5Store(filename, **kwargs) as h5s:
        list(h5s)


class TestH5StoreMultiProcessing(TestH5StoreBase):
    def test_single_writer_multiple_reader_same_process(self):
        with self.open_h5store() as writer:
            with self.open_h5store():  # second writer
                with self.open_h5store(mode="r") as reader1:
                    with self.open_h5store(mode="r") as reader2:
                        writer["test"] = True
                        assert writer["test"]
                        assert reader1["test"]
                        assert reader2["test"]

    @pytest.mark.skipif(
        python_implementation() != "CPython", reason="SWMR mode not available."
    )
    def test_single_writer_multiple_reader_same_instance(self):
        from multiprocessing import Process

        def read():
            p = Process(
                target=_read_from_h5store,
                args=(self._fn_store,),
                kwargs=(dict(mode="r", swmr=True, libver="latest")),
            )
            p.start()
            p.join()
            # Ensure the process succeeded
            assert p.exitcode == 0

        with self.open_h5store(libver="latest") as writer:
            writer.file.swmr_mode = True
            read()
            writer["test"] = True
            read()

    def test_multiple_reader_different_process_no_swmr(self):
        read_cmd = (
            r'python -c "from signac.h5store import H5Store; '
            r"h5s = H5Store({}, mode=\"r\"); list(h5s); "
            r'h5s.close()"'
        ).format(repr(self._fn_store))

        with self.open_h5store():
            pass  # create file

        try:
            with self.open_h5store(mode="r"):  # single reader
                subprocess.check_output(read_cmd, shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as error:
            print("\n", error.output.decode(), file=sys.stderr)
            raise

    def test_single_writer_multiple_reader_different_process_no_swmr(self):
        read_cmd = (
            r'python -c "from signac.h5store import H5Store; '
            r"h5s = H5Store({}, mode=\"r\"); list(h5s); "
            r'h5s.close()"'
        ).format(repr(self._fn_store))

        with self.open_h5store():  # single writer
            with pytest.raises(subprocess.CalledProcessError):
                subprocess.check_output(read_cmd, shell=True, stderr=subprocess.DEVNULL)

    @pytest.mark.skipif(
        python_implementation() != "CPython", reason="SWMR mode not available."
    )
    def test_single_writer_multiple_reader_different_process_swmr(self):
        read_cmd = (
            r'python -c "from signac.h5store import H5Store; '
            r"h5s = H5Store({}, mode=\"r\", swmr=True); list(h5s); "
            r'h5s.close()"'
        ).format(repr(self._fn_store))

        with self.open_h5store(libver="latest") as writer:
            with pytest.raises(subprocess.CalledProcessError):
                subprocess.check_output(read_cmd, shell=True, stderr=subprocess.DEVNULL)

        try:
            with self.open_h5store(libver="latest") as writer:
                writer.file.swmr_mode = True
                subprocess.check_output(read_cmd, shell=True, stderr=subprocess.STDOUT)
                writer["test"] = True
                subprocess.check_output(read_cmd, shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as error:
            print("\n", error.output.decode(), file=sys.stderr)
            raise


@pytest.mark.skipif(not NUMPY, reason="requires numpy package")
@pytest.mark.skipif(
    python_implementation() != "CPython", reason="Optimized for CPython."
)
class TestH5StorePerformance(TestH5StoreBase):
    max_slowdown_vs_native_factor = 1.25

    @pytest.fixture
    def setUp(self, setUp_base_h5Store):
        value = self.get_testdata()
        times = numpy.zeros(200)
        for i in range(len(times)):
            start = time()
            with h5py.File(self._fn_store, mode="a") as h5file:
                if i:
                    del h5file["_baseline"]
                h5file.create_dataset("_baseline", data=value, shape=None)
            times[i] = time() - start
        self.baseline_time = times

    def assertSpeed(self, times):
        msg = "\n{:>10}\t{:>8}\t{:>8}\t{:>4}\n".format(
            "", "Measurement", "Benchmark", "Factor"
        )

        def format_row(text, reducer):
            return "{:<10}\t{:.2e}\t{:.2e}\t{:.3}\n".format(
                text,
                reducer(times),
                reducer(self.baseline_time),
                reducer(times) / reducer(self.baseline_time),
            )

        msg += format_row("mean", numpy.mean)
        msg += format_row("median", numpy.median)
        msg += format_row("25 percentile", partial(numpy.percentile, q=25))
        msg += format_row("75 percentile", partial(numpy.percentile, q=75))
        assert (
            numpy.percentile(times, 25) / numpy.percentile(self.baseline_time, 75)
            < self.max_slowdown_vs_native_factor
        ), msg

    @pytest.mark.skipif(
        WINDOWS, reason="This test fails for an unknown reason on Windows."
    )
    @pytest.mark.skip(
        reason="This test fails randomly on CI. "
        "See https://github.com/glotzerlab/signac/pull/307"
    )
    def test_speed_get(self, setUp):
        times = numpy.zeros(200)
        key = "test_speed_get"
        value = self.get_testdata()
        self.get_h5store()[key] = value
        self.assertEqual(self.get_h5store()[key], value)  # sanity check
        for i in range(len(times)):
            start = time()
            self.get_h5store()[key]
            times[i] = time() - start
        self.assertSpeed(times)

    @pytest.mark.skipif(
        WINDOWS, reason="This test fails for an unknown reason on Windows."
    )
    @pytest.mark.skip(
        reason="This test fails randomly on CI. "
        "See https://github.com/glotzerlab/signac/pull/307"
    )
    def test_speed_set(self, setUp):
        times = numpy.zeros(200)
        key = "test_speed_set"
        value = self.get_testdata()
        for i in range(len(times)):
            start = time()
            self.get_h5store()[key] = value
            times[i] = time() - start
        self.assertEqual(self.get_h5store()[key], value)  # sanity check
        self.assertSpeed(times)


class TestH5StorePerformanceNestedData(TestH5StorePerformance):
    max_slowdown_vs_native_factor = 1.75

    def get_testdata(self, size=None):
        return dict(a=super().get_testdata(size))

    @pytest.fixture
    def setUp(self, setUp_base_h5Store):
        value = TestH5StorePerformance.get_testdata(self)
        times = numpy.zeros(200)
        for i in range(len(times)):
            start = time()
            with h5py.File(self._fn_store, mode="a") as h5file:
                if i:
                    del h5file["_basegroup"]
                h5file.create_group("_basegroup").create_dataset(
                    "_baseline", data=value, shape=None
                )
            times[i] = time() - start
        self.baseline_time = times



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Utility functions."""

import os.path
import sys
from collections.abc import Mapping


def _print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def _safe_relpath(path):
    """Attempt to make a relative path, or return the original path.

    This is useful for logging and representing objects, where an absolute path
    may be very long.
    """
    try:
        return os.path.relpath(path)
    except ValueError:
        # Windows cannot find relative paths across drives, so show the
        # original path instead.
        return path


def _query_yes_no(question, default="yes"):  # pragma: no cover
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
    It must be "yes" (the default), "no" or None (meaning
    an answer is required of the user).

    The "answer" return value is one of "yes" or "no".

    Parameters
    ----------
    question : str
        Question presented to the user.
    default : str, optional
        Presumed answer if the user just hits <Enter> (Default value = "yes").

    Returns
    -------
    bool
        ``True`` if yes, ``False`` if no.

    Raises
    ------
    ValueError
        When default is set to invalid answer.

    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        choice = input(question + prompt).lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' (or 'y' or 'n').")


def _mkdir_p(path):
    """Make a new directory, or do nothing if the directory already exists.

    Parameters
    ----------
    path : str
        New directory name.

    """
    # Performance: `isdir` is fast and eliminates the rest of os.makedirs
    # if the path already exists. Typically this function is called in cases
    # where the path already exists. If the path is a file, this check returns
    # False and allows os.makedirs to raise FileExistsError as usual.
    if not os.path.isdir(path):
        os.makedirs(path, exist_ok=True)


def _dotted_dict_to_nested_dicts(dotted_dict, delimiter_nested="."):
    """Convert dotted keys in the state point dict to a nested dict.

    Parameters
    ----------
    dotted_dict : dict
        A mapping with dots/delimiter_nested in its keys, e.g. {'a.b': 'c'}.
    delimiter_nested : str
        A string delimiter between keys, defaults to '.'.

    Returns
    -------
    dict
        A mapping instance with nested dicts, e.g. {'a': {'b': 'c'}}.

    """
    nested_dict = {}
    for key, value in dotted_dict.items():
        tokens = key.split(delimiter_nested)
        if len(tokens) > 1:
            tmp = nested_dict.setdefault(tokens[0], {})
            for token in tokens[1:-1]:
                tmp = tmp.setdefault(token, {})
            tmp[tokens[-1]] = value
        else:
            nested_dict[tokens[0]] = value
    return nested_dict


class _hashable_dict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def _to_hashable(obj):
    """Create a hash of passed type.

    Parameters
    ----------
    obj
        Object to create a hashable version of. Lists are converted
        to tuples, and hashes are defined for dicts.

    Returns
    -------
    Hash created for obj.

    """
    if type(obj) is list:
        return tuple(_to_hashable(_) for _ in obj)
    elif type(obj) is dict:
        return _hashable_dict(obj)
    else:
        return obj


def _nested_dicts_to_dotted_keys(d, key=None):
    """Generate tuples of key in dotted string format and value from nested dict.

    Parameters
    ----------
    d : dict
        A mapping instance with nested dicts, e.g. {'a': {'b': 'c'}}.
    key : str
        Key of root at current point in the recursion, used to
        build up nested keys in the top-level dict through
        multiple recursive calls (Default value = None).

    Yields
    ------
    tuple
        Tuples of dotted key and values e.g. ('a.b', 'c')

    """
    if isinstance(d, Mapping):
        if d:
            for k in d:
                k_ = k if key is None else ".".join((key, k))
                yield from _nested_dicts_to_dotted_keys(d[k], key=k_)
        elif key is not None:
            yield key, d
    else:
        if type(d) is list:
            d = _to_hashable(d)
        yield key, d



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import copy
import json
import logging
import os
import random
import uuid
from contextlib import contextmanager
from tempfile import TemporaryDirectory

import pytest

import signac
from signac._config import _load_config
from signac.errors import (
    DestinationExistsError,
    InvalidKeyError,
    JobsCorruptedError,
    KeyTypeError,
)
from signac.job import Job

try:
    import h5py  # noqa: F401

    H5PY = True
except ImportError:
    H5PY = False

# Make sure the jobs created for this test are unique.
test_token = {"test_token": str(uuid.uuid4())}

BUILTINS = [
    ({"e": [1.0, "1.0", 1, True]}, "4d8058a305b940005be419b30e99bb53"),
    ({"d": True}, "33cf9999de25a715a56339c6c1b28b41"),
    ({"f": (1.0, "1.0", 1, True)}, "e998db9b595e170bdff936f88ccdbf75"),
    ({"a": 1}, "42b7b4f2921788ea14dac5566e6f06d0"),
    ({"c": "1.0"}, "80fa45716dd3b83fa970877489beb42e"),
    ({"b": 1.0}, "0ba6c5a46111313f11c41a6642520451"),
]


def builtins_dict():
    random.shuffle(BUILTINS)
    d = {}
    for b in BUILTINS:
        d.update(b[0])
    return d


BUILTINS_HASH = "7a80b58db53bbc544fc27fcaaba2ce44"

NESTED_HASH = "bd6f5828f4410b665bffcec46abeb8f3"


class TestJobBase:
    project_class = signac.Project

    @pytest.fixture(autouse=True)
    def setUp(self, request):
        self._tmp_dir = TemporaryDirectory(prefix="signac_")
        request.addfinalizer(self._tmp_dir.cleanup)
        self._tmp_pr = os.path.join(self._tmp_dir.name, "pr")
        os.mkdir(self._tmp_pr)
        self.config = _load_config()
        self.project = self.project_class.init_project(path=self._tmp_pr)

    def tearDown(self):
        pass

    def open_job(self, *args, **kwargs):
        project = self.project
        return project.open_job(*args, **kwargs)

    @classmethod
    def nested_dict(self):
        d = dict(builtins_dict())
        d["g"] = builtins_dict()
        return d


class TestJobID(TestJobBase):
    def test_builtins(self):
        for p, h in BUILTINS:
            assert str(self.project.open_job(p)) == h
        assert str(self.project.open_job(builtins_dict())) == BUILTINS_HASH

    def test_shuffle(self):
        for i in range(10):
            assert str(self.project.open_job(builtins_dict())) == BUILTINS_HASH

    def test_nested(self):
        for i in range(10):
            assert str(self.project.open_job(self.nested_dict())) == NESTED_HASH

    def test_sequences_identity(self):
        job1 = self.project.open_job({"a": [1.0, "1.0", 1, True]})
        job2 = self.project.open_job({"a": (1.0, "1.0", 1, True)})
        assert str(job1) == str(job2)
        assert job1.statepoint() == job2.statepoint()


class TestJob(TestJobBase):
    def test_repr(self):
        job = self.project.open_job({"a": 0})
        job2 = self.project.open_job({"a": 0})
        assert repr(job) == repr(job2)
        assert job == job2

    def test_str(self):
        job = self.project.open_job({"a": 0})
        assert str(job) == job.id

    def test_eq(self):
        job = self.project.open_job({"a": 0})

        # Make sure that Jobs can only be equal to other Job instances.
        class NonJob:
            """Minimal class that cannot be compared with Job objects."""

            def __init__(self, job):
                self.id = job.id
                self._workspace = job.path

        class JobSubclass(Job):
            """Minimal subclass that can be compared with Job objects."""

            def __init__(self, job):
                self._id = job.id
                self._path = job.path

            def workspace(self):
                return self._workspace

        non_job = NonJob(job)
        assert job != non_job
        assert non_job != job

        sub_job = JobSubclass(job)
        assert job == sub_job
        assert sub_job == job

        job2 = self.project.open_job({"a": 0})
        assert job == job2
        assert job2 == job

    def test_isfile(self):
        job = self.project.open_job({"a": 0})
        fn = "test.txt"
        fn_ = os.path.join(job.path, fn)
        assert not job.isfile(fn)
        job.init()
        assert not job.isfile(fn)
        with open(fn_, "w") as file:
            file.write("hello")
        assert job.isfile(fn)

    def test_copy(self):
        job = self.project.open_job({"a": 0}).init()
        assert job in self.project

        # Modify copy
        copied_job = copy.copy(job)
        assert job is not copied_job
        assert job == copied_job
        assert job.sp == copied_job.sp
        assert job in self.project
        assert copied_job in self.project
        copied_job.sp.a = 1
        assert job in self.project
        assert copied_job in self.project
        assert job == copied_job
        assert job.sp == copied_job.sp

        # Modify original
        copied_job = copy.copy(job)
        assert job is not copied_job
        assert job == copied_job
        assert job.sp == copied_job.sp
        assert job in self.project
        assert copied_job in self.project
        job.sp.a = 2
        assert job in self.project
        assert copied_job in self.project
        assert job == copied_job
        assert job.sp == copied_job.sp

        # Delete original
        del job
        assert copied_job in self.project
        copied_job.sp.a = 3
        assert copied_job in self.project

    def test_deepcopy(self):
        job = self.project.open_job({"a": 0}).init()
        assert job in self.project

        # Modify copy
        copied_job = copy.deepcopy(job)
        assert job is not copied_job
        assert job == copied_job
        assert job.sp == copied_job.sp
        assert job in self.project
        assert copied_job in self.project
        copied_job.sp.a = 1
        assert job not in self.project
        assert copied_job in self.project
        assert job != copied_job
        assert job.sp != copied_job.sp

        # Modify original
        job = self.project.open_job({"a": 0}).init()
        copied_job = copy.deepcopy(job)
        assert job is not copied_job
        assert job == copied_job
        assert job.sp == copied_job.sp
        assert job in self.project
        assert copied_job in self.project
        job.sp.a = 2
        assert job in self.project
        assert copied_job not in self.project
        assert job != copied_job
        assert job.sp != copied_job.sp

        # Delete original
        copied_job = copy.deepcopy(job)
        del job
        assert copied_job in self.project
        copied_job.sp.a = 3
        assert copied_job in self.project

    def test_project_access_from_job(self):
        job = self.project.open_job({"a": 0}).init()
        assert isinstance(job.project, signac.Project)
        assert job in job.project
        assert job.project.path == self._tmp_pr

    def test_custom_project_access_from_job(self):
        # Test a custom project subclass to ensure compatibility with signac-flow's FlowProject
        class CustomProject(signac.Project):
            pass

        custom_project = CustomProject.get_project(self._tmp_pr)
        job = custom_project.open_job({"a": 0}).init()
        assert isinstance(job.project, CustomProject)
        assert job in job.project
        assert job.project.path == self._tmp_pr


class TestJobSpInterface(TestJobBase):
    def test_interface_read_only(self):
        sp = self.nested_dict()
        job = self.open_job(sp)
        assert job.statepoint() == json.loads(json.dumps(sp))
        for x in ("a", "b", "c", "d", "e"):
            assert getattr(job.sp, x) == sp[x]
            assert job.sp[x] == sp[x]
        for x in ("a", "b", "c", "d", "e"):
            assert getattr(job.sp.g, x) == sp["g"][x]
            assert job.sp[x] == sp[x]
        for x in ("a", "b", "c", "d", "e"):
            assert job.sp.get(x) == sp[x]
            assert job.sp.get(x) == sp[x]
            assert job.sp.g.get(x) == sp["g"][x]
        assert job.sp.get("not_in_sp") is None
        assert job.sp.g.get("not_in_sp") is None
        assert job.sp.get("not_in_sp", None) is None
        assert job.sp.g.get("not_in_sp", None) is None
        assert job.sp.get("not_in_sp", 23) == 23
        assert job.sp.g.get("not_in_sp", 23) == 23

    def test_interface_contains(self):
        sp = self.nested_dict()
        job = self.open_job(sp)
        for x in ("a", "b", "c", "d", "e"):
            assert x in job.sp
            assert x in job.sp.g

    def test_interface_read_write(self):
        sp = self.nested_dict()
        job = self.open_job(sp)
        job.init()
        for x in ("a", "b", "c", "d", "e"):
            assert getattr(job.sp, x) == sp[x]
            assert job.sp[x] == sp[x]
        for x in ("a", "b", "c", "d", "e"):
            assert getattr(job.sp.g, x) == sp["g"][x]
            assert job.sp[x] == sp[x]
        a = [1, 1.0, "1.0", True, None]
        for x in ("a", "b", "c", "d", "e"):
            setattr(job.sp, x, a)
            assert getattr(job.sp, x) == a
            setattr(job.sp.g, x, a)
            assert getattr(job.sp.g, x) == a
        t = (1, 2, 3)  # tuple
        job.sp.t = t
        assert job.sp.t == list(t)  # implicit conversion
        job.sp.g.t = t
        assert job.sp.g.t == list(t)

    def test_interface_job_identity_change(self):
        job = self.open_job({"a": 0})
        old_id = job.id
        job.sp.a = 1
        assert old_id != job.id

    def test_interface_nested_kws(self):
        with pytest.raises(InvalidKeyError):
            job = self.open_job({"a.b.c": 0})

        job = self.open_job(dict(a=dict(b=dict(c=2))))
        assert job.sp.a.b.c == 2
        assert job.sp["a"]["b"]["c"] == 2

    def test_interface_lists(self):
        job = self.open_job({"a": [1, 2, 3]})
        assert job.sp.a == [1, 2, 3]
        old_id = job.id
        job.sp.a.append(4)
        assert job.sp.a == [1, 2, 3, 4]
        assert old_id != job.id

    def test_interface_reserved_keywords(self):
        job = self.open_job({"with": 0, "pop": 1})
        assert job.sp["with"] == 0
        assert job.sp["pop"] == 1
        assert job.sp.pop("with") == 0
        assert "with" not in job.sp

    def test_interface_illegal_type(self):
        job = self.open_job(dict(a=0))
        assert job.sp.a == 0

        class Foo:
            pass

        with pytest.raises(TypeError):
            job.sp.a = Foo()

    def test_interface_rename(self):
        job = self.open_job(dict(a=0))
        job.init()
        assert job.sp.a == 0
        job.sp.b = job.sp.pop("a")
        assert "a" not in job.sp
        assert job.sp.b == 0

    def test_interface_copy(self):
        job = self.open_job(dict(a=0)).init()
        copy.copy(job.sp).a = 1
        assert job in self.project

    def test_interface_deepcopy(self):
        job = self.open_job(dict(a=0)).init()
        copy.deepcopy(job.sp).a = 1
        assert job not in self.project

    def test_interface_add(self):
        job = self.open_job(dict(a=0))
        job.init()
        with pytest.raises(AttributeError):
            job.sp.b
        job.sp.b = 1
        assert "b" in job.sp
        assert job.sp.b == 1

    def test_interface_delete(self):
        job = self.open_job(dict(a=0, b=0))
        job.init()
        assert "b" in job.sp
        assert job.sp.b == 0
        del job.sp["b"]
        assert "b" not in job.sp
        with pytest.raises(AttributeError):
            job.sp.b
        job.sp.b = 0
        assert "b" in job.sp
        assert job.sp.b == 0
        del job.sp.b
        assert "b" not in job.sp
        with pytest.raises(AttributeError):
            job.sp.b

    def test_interface_destination_conflict(self):
        job_a = self.open_job(dict(a=0))
        job_b = self.open_job(dict(b=0))
        job_a.init()
        id_a = job_a.id
        job_a.sp = dict(b=0)
        assert job_a.statepoint() == dict(b=0)
        assert job_a == job_b
        assert job_a.id != id_a
        job_a = self.open_job(dict(a=0))
        # Moving to existing job, no problem while empty:
        assert job_a != job_b
        job_a.sp = dict(b=0)
        job_a = self.open_job(dict(a=0))
        job_b.init()
        # Moving to an existing job with data leads
        # to an error:
        job_a.document["a"] = 0
        job_b.document["a"] = 0
        assert job_a != job_b
        with pytest.raises(RuntimeError):
            job_a.sp = dict(b=0)
        with pytest.raises(DestinationExistsError):
            job_a.sp = dict(b=0)

    def test_interface_multiple_changes(self):
        for i in range(1, 4):
            job = self.project.open_job(dict(a=i))
            job.init()
        for job in self.project:
            assert job.sp.a > 0

        for job in self.project:
            obj_id = id(job)
            id0 = job.id
            sp0 = job.statepoint()
            assert id(job) == obj_id
            assert job.sp.a > 0
            assert job.id == id0
            assert job.sp == sp0

            job.sp.a = -job.sp.a
            assert id(job) == obj_id
            assert job.sp.a < 0
            assert job.id != id0
            assert job.sp != sp0

            job.sp.a = -job.sp.a
            assert id(job) == obj_id
            assert job.sp.a > 0
            assert job.id == id0
            assert job.sp == sp0
            job2 = self.project.open_job(id=id0)
            assert job.sp == job2.sp
            assert job.id == job2.id

    def test_valid_sp_key_types(self):
        job = self.open_job(dict(valid_key=True)).init()

        # Only strings are permitted as keys
        for key in ("0", "1"):
            job.sp[key] = "test"
            assert str(key) in job.sp

    def test_invalid_sp_key_types(self):
        class A:
            pass

        with pytest.raises(KeyTypeError):
            self.open_job({A(): True}).init()

        job = self.open_job(dict(invalid_key=True)).init()

        for key in (1, True, False, None, 0.0, A(), (1, 2, 3)):
            with pytest.raises(KeyTypeError):
                job.sp[key] = "test"
            with pytest.raises(KeyTypeError):
                job.sp = {key: "test"}
        for key in ([], {}):
            with pytest.raises(TypeError):
                job.sp[key] = "test"
            with pytest.raises(TypeError):
                job.sp = {key: "test"}

    def test_valid_doc_key_types(self):
        job = self.open_job(dict(valid_key=True)).init()

        # Only strings are permitted as keys
        for key in ("0", "1"):
            job.doc[key] = "test"
            assert str(key) in job.doc

    def test_invalid_doc_key_types(self):
        job = self.open_job(dict(invalid_key=True)).init()

        class A:
            pass

        for key in (1, True, False, None, 0.0, A(), (1, 2, 3)):
            with pytest.raises(KeyTypeError):
                job.doc[key] = "test"
            with pytest.raises(KeyTypeError):
                job.doc = {key: "test"}
        for key in ([], {}):
            with pytest.raises(TypeError):
                job.doc[key] = "test"
            with pytest.raises(TypeError):
                job.doc = {key: "test"}


class TestConfig(TestJobBase):
    def test_config_str(self):
        str(self.project.config)


class TestJobOpenAndClosing(TestJobBase):
    def test_init(self):
        job = self.open_job(test_token)
        assert not os.path.isdir(job.path)
        job.init()
        assert os.path.isdir(job.path)
        assert os.path.exists(os.path.join(job.path, job.FN_STATE_POINT))

    def test_chained_init(self):
        job = self.open_job(test_token)
        assert not os.path.isdir(job.path)
        job = self.open_job(test_token).init()
        assert os.path.isdir(job.path)
        assert os.path.exists(os.path.join(job.path, job.FN_STATE_POINT))

    def test_construction(self):
        from signac import Project  # noqa: F401

        # The eval statement needs to have Project available
        job = self.open_job(test_token)
        job2 = eval(repr(job))
        assert job == job2

    def test_open_job_close(self):
        with self.open_job(test_token) as job:
            pass
        job.remove()

    def test_open_job_close_manual(self):
        job = self.open_job(test_token)
        job.open()
        job.close()
        job.remove()

    def test_open_job_close_with_error(self):
        job = self.open_job(test_token)

        class TestError(Exception):
            pass

        with pytest.raises(TestError):
            with job:
                raise TestError()
        job.remove()

    def test_reopen_job(self):
        with self.open_job(test_token) as job:
            job_id = job.id
            assert str(job_id) == str(job)

        with self.open_job(test_token) as job:
            assert job.id == job_id
        job.remove()

    def test_close_nonopen_job(self):
        job = self.open_job(test_token)
        job.close()
        with job:
            pass

    def test_close_job_while_open(self):
        rp = os.path.realpath
        cwd = rp(os.getcwd())
        job = self.open_job(test_token)
        with job:
            job.close()
            assert cwd == rp(os.getcwd())

    def test_open_job_recursive(self):
        rp = os.path.realpath
        cwd = rp(os.getcwd())
        job = self.open_job(test_token)
        with job:
            assert rp(job.path) == rp(os.getcwd())
        assert cwd == rp(os.getcwd())
        with job:
            assert rp(job.path) == rp(os.getcwd())
            os.chdir(self.project.path)
        assert cwd == rp(os.getcwd())
        with job:
            assert rp(job.path) == rp(os.getcwd())
            with job:
                assert rp(job.path) == rp(os.getcwd())
            assert rp(job.path) == rp(os.getcwd())
        assert cwd == rp(os.getcwd())
        with job:
            assert rp(job.path) == rp(os.getcwd())
            os.chdir(self.project.path)
            with job:
                assert rp(job.path) == rp(os.getcwd())
            assert rp(os.getcwd()) == rp(self.project.path)
        assert cwd == rp(os.getcwd())
        with job:
            job.close()
            assert cwd == rp(os.getcwd())
            with job:
                assert rp(job.path) == rp(os.getcwd())
            assert cwd == rp(os.getcwd())
        assert cwd == rp(os.getcwd())

    def test_corrupt_workspace(self):
        job = self.open_job(test_token)
        job.init()
        fn_statepoint = os.path.join(job.path, job.FN_STATE_POINT)
        with open(fn_statepoint, "w") as file:
            file.write("corrupted")
        job2 = self.open_job(test_token)
        try:
            logging.disable(logging.ERROR)
            with pytest.raises(JobsCorruptedError):
                job2.init()
        finally:
            logging.disable(logging.NOTSET)
        job2.init(force=True)
        job2.init()


class TestJobDocument(TestJobBase):
    def test_get_set(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        assert not bool(job.document)
        assert len(job.document) == 0
        assert key not in job.document
        job.document[key] = d
        assert bool(job.document)
        assert len(job.document) == 1
        assert key in job.document
        assert job.document[key] == d
        assert job.document.get(key) == d
        assert job.document.get("non-existent-key", d) == d

    def test_del(self):
        key0 = "del0"
        key1 = "del1"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        assert len(job.document) == 0
        assert key0 not in job.document
        job.document[key0] = d0
        assert len(job.document) == 1
        assert key0 in job.document
        job.document[key1] = d1
        assert len(job.document) == 2
        assert key0 in job.document
        assert key1 in job.document
        assert job.document[key0] == d0
        assert job.document[key1] == d1
        del job.document[key0]
        assert len(job.document) == 1
        assert key1 in job.document
        assert key0 not in job.document

    def test_get_set_doc(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        assert not bool(job.doc)
        assert len(job.doc) == 0
        assert key not in job.doc
        job.doc[key] = d
        assert bool(job.doc)
        assert len(job.doc) == 1
        assert key in job.doc
        assert job.doc[key] == d
        assert job.doc.get(key) == d
        assert job.doc.get("non-existent-key", d) == d

    def test_set_set_doc(self):
        key0 = "set_set0"
        key1 = "set_set1"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        assert not bool(job.doc)
        assert len(job.doc) == 0
        assert key0 not in job.doc
        job.doc[key0] = d0
        assert bool(job.doc)
        assert len(job.doc) == 1
        assert key0 in job.doc
        assert job.doc[key0] == d0
        job = self.open_job(test_token)
        assert bool(job.doc)
        assert len(job.doc) == 1
        assert key0 in job.doc
        assert job.doc[key0] == d0
        job = self.open_job(test_token)
        job.document[key1] = d1
        assert bool(job.doc)
        assert len(job.doc) == 2
        assert key0 in job.doc
        assert key1 in job.doc
        assert job.doc[key0] == d0
        assert job.doc[key1] == d1

    def test_get_set_nested(self):
        d0 = "test_data0"
        d1 = "test_data1"
        d2 = "test_data2"
        assert d0 != d1 != d2
        job = self.open_job(test_token)
        assert len(job.document) == 0
        assert "key0" not in job.document
        job.document["key0"] = d0
        assert len(job.document) == 1
        assert "key0" in job.document
        assert job.document["key0"] == d0
        with pytest.raises(AttributeError):
            job.document.key0.key1
        job.document.key0 = {"key1": d0}
        assert len(job.document) == 1
        assert "key0" in job.document
        assert job.document() == {"key0": {"key1": d0}}
        assert job.document["key0"] == {"key1": d0}
        assert job.document["key0"]["key1"] == d0
        assert job.document.key0 == {"key1": d0}
        assert job.document.key0.key1 == d0
        job.document.key0.key1 = d1
        assert job.document == {"key0": {"key1": d1}}
        assert job.document["key0"] == {"key1": d1}
        assert job.document["key0"]["key1"] == d1
        assert job.document.key0 == {"key1": d1}
        assert job.document.key0.key1 == d1
        job.document["key0"]["key1"] = d2
        assert job.document == {"key0": {"key1": d2}}
        assert job.document["key0"] == {"key1": d2}
        assert job.document["key0"]["key1"] == d2
        assert job.document.key0 == {"key1": d2}
        assert job.document.key0.key1 == d2

    def test_get_set_nested_doc(self):
        d0 = "test_data0"
        d1 = "test_data1"
        d2 = "test_data2"
        assert d0 != d1 != d2
        job = self.open_job(test_token)
        assert len(job.doc) == 0
        assert "key0" not in job.doc
        job.doc["key0"] = d0
        assert len(job.doc) == 1
        assert "key0" in job.doc
        assert job.doc["key0"] == d0
        with pytest.raises(AttributeError):
            job.doc.key0.key1
        job.doc.key0 = {"key1": d0}
        assert len(job.doc) == 1
        assert "key0" in job.doc
        assert job.doc() == {"key0": {"key1": d0}}
        assert job.doc["key0"] == {"key1": d0}
        assert job.doc["key0"]["key1"] == d0
        assert job.doc.key0 == {"key1": d0}
        assert job.doc.key0.key1 == d0
        job.doc.key0.key1 = d1
        assert job.doc == {"key0": {"key1": d1}}
        assert job.doc["key0"] == {"key1": d1}
        assert job.doc["key0"]["key1"] == d1
        assert job.doc.key0 == {"key1": d1}
        assert job.doc.key0.key1 == d1
        job.doc["key0"]["key1"] = d2
        assert job.doc == {"key0": {"key1": d2}}
        assert job.doc["key0"] == {"key1": d2}
        assert job.doc["key0"]["key1"] == d2
        assert job.doc.key0 == {"key1": d2}
        assert job.doc.key0.key1 == d2

    def test_assign(self):
        key = "assign"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        assert len(job.document) == 0
        job.document[key] = d0
        assert len(job.document) == 1
        assert job.document() == {key: d0}
        with pytest.raises(ValueError):
            job.document = d1
        job.document = {key: d1}
        assert len(job.document) == 1
        assert job.document() == {key: d1}

    def test_assign_doc(self):
        key = "assign"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        assert len(job.doc) == 0
        job.doc[key] = d0
        assert len(job.doc) == 1
        assert job.doc() == {key: d0}
        with pytest.raises(ValueError):
            job.doc = d1
        job.doc = {key: d1}
        assert len(job.doc) == 1
        assert job.doc() == {key: d1}

    def test_copy_document(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        job.document[key] = d
        assert bool(job.document)
        assert len(job.document) == 1
        assert key in job.document
        assert job.document[key] == d
        assert job.document.get(key) == d
        assert job.document.get("non-existent-key", d) == d
        copy = dict(job.document)
        assert bool(copy)
        assert len(copy) == 1
        assert key in copy
        assert copy[key] == d
        assert copy.get(key) == d
        assert copy.get("non-existent-key", d) == d

    def test_update(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        job.document.update({key: d})
        assert key in job.document

    def test_clear_document(self):
        key = "clear"
        d = "test_data"
        job = self.open_job(test_token)
        job.document[key] = d
        assert key in job.document
        assert len(job.document) == 1
        job.document.clear()
        assert key not in job.document
        assert len(job.document) == 0

    def test_reopen(self):
        key = "clear"
        d = "test_data"
        job = self.open_job(test_token)
        job.document[key] = d
        assert key in job.document
        assert len(job.document) == 1
        job2 = self.open_job(test_token)
        assert key in job2.document
        assert len(job2.document) == 1

    def test_concurrency(self):
        key = "concurrent"
        d = "test_data"
        job = self.open_job(test_token)
        job2 = self.open_job(test_token)
        assert key not in job.document
        assert key not in job2.document
        job.document[key] = d
        assert key in job.document
        assert key in job2.document

    def test_remove(self):
        key = "remove"
        job = self.open_job(test_token)
        job.remove()
        d = "test_data"
        job.document[key] = d
        assert key in job.document
        assert len(job.document) == 1
        fn_test = os.path.join(job.path, "test")
        with open(fn_test, "w") as file:
            file.write("test")
        assert os.path.isfile(fn_test)
        job.remove()
        assert key not in job.document
        assert not os.path.isfile(fn_test)

    def test_clear_job(self):
        key = "clear"
        job = self.open_job(test_token)
        assert job not in self.project
        job.clear()
        assert job not in self.project
        job.clear()
        assert job not in self.project
        job.init()
        assert job in self.project
        job.clear()
        assert job in self.project
        job.clear()
        job.clear()
        assert job in self.project
        d = "test_data"
        job.document[key] = d
        assert job in self.project
        assert key in job.document
        assert len(job.document) == 1
        job.clear()
        assert len(job.document) == 0
        with open(job.fn("test"), "w") as file:
            file.write("test")
        assert job.isfile("test")
        assert job in self.project
        job.clear()
        assert not job.isfile("test")
        assert len(job.document) == 0

    def test_reset(self):
        key = "reset"
        job = self.open_job(test_token)
        assert job not in self.project
        job.reset()
        assert job in self.project
        assert len(job.document) == 0
        job.document[key] = "test_data"
        assert len(job.document) == 1
        job.reset()
        assert job in self.project
        assert len(job.document) == 0

    def test_doc(self):
        key = "test_doc"
        job = self.open_job(test_token)

        def check_content(key, d):
            assert job.doc[key] == d
            assert getattr(job.doc, key) == d
            assert job.doc()[key] == d
            assert job.document[key] == d
            assert getattr(job.document, key) == d
            assert job.document()[key] == d

        d0 = "test_data0"
        job.doc[key] = d0
        check_content(key, d0)
        d1 = "test_data1"
        job.doc[key] = d1
        check_content(key, d1)
        d2 = "test_data2"
        job.document[key] = d2
        check_content(key, d2)
        d3 = "test_data3"
        setattr(job.doc, key, d3)
        check_content(key, d3)

    def test_sp_formatting(self):
        job = self.open_job({"a": 0})
        assert f"{job.statepoint.a}" == str(job.sp.a)
        assert f"{job.sp.a}" == str(job.sp.a)
        assert "{job.statepoint[a]}".format(job=job) == str(job.sp.a)
        assert "{job.sp[a]}".format(job=job) == str(job.sp.a)
        job.sp.a = dict(b=0)
        assert f"{job.statepoint.a.b}" == str(job.sp.a.b)
        assert f"{job.sp.a.b}" == str(job.sp.a.b)
        assert "{job.statepoint[a][b]}".format(job=job) == str(job.sp.a.b)
        assert "{job.sp[a][b]}".format(job=job) == str(job.sp.a.b)

    def test_doc_formatting(self):
        job = self.open_job(test_token)
        job.doc.a = 0
        assert f"{job.doc.a}" == str(job.doc.a)
        assert "{job.doc[a]}".format(job=job) == str(job.doc.a)
        assert f"{job.document.a}" == str(job.doc.a)
        assert "{job.document[a]}".format(job=job) == str(job.doc.a)
        job.doc.a = dict(b=0)
        assert f"{job.doc.a.b}" == str(job.doc.a.b)
        assert f"{job.doc.a.b}" == str(job.doc.a.b)
        assert f"{job.document.a.b}" == str(job.doc.a.b)
        assert "{job.document[a][b]}".format(job=job) == str(job.doc.a.b)

    @pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
    def test_reset_statepoint_job(self):
        key = "move_job"
        d = "test_data"
        src = test_token
        dst = dict(test_token)
        dst["dst"] = True
        src_job = self.open_job(src)
        src_job.document[key] = d
        assert key in src_job.document
        assert len(src_job.document) == 1
        src_job.data[key] = d
        assert key in src_job.data
        assert len(src_job.data) == 1
        src_job.statepoint = dst
        src_job = self.open_job(src)
        dst_job = self.open_job(dst)
        assert key in dst_job.document
        assert len(dst_job.document) == 1
        assert key not in src_job.document
        assert key in dst_job.data
        assert len(dst_job.data) == 1
        assert key not in src_job.data
        with pytest.raises(RuntimeError):
            src_job.statepoint = dst
        with pytest.raises(DestinationExistsError):
            src_job.statepoint = dst

    @pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
    def test_reset_statepoint_job_lazy_access(self):
        key = "move_job"
        d = "test_data"
        src = test_token
        dst = dict(test_token)
        dst["dst"] = True
        src_job = self.open_job(src)
        src_job.document[key] = d
        assert key in src_job.document
        assert len(src_job.document) == 1
        src_job.data[key] = d
        assert key in src_job.data
        assert len(src_job.data) == 1
        # Clear the project's state point cache to force lazy load
        self.project._sp_cache.clear()
        src_job_by_id = self.open_job(id=src_job.id)
        # Check that the state point will be instantiated lazily during the
        # call to reset_statepoint
        assert src_job_by_id._statepoint_requires_init
        src_job_by_id.statepoint = dst
        src_job = self.open_job(src)
        dst_job = self.open_job(dst)
        assert key in dst_job.document
        assert len(dst_job.document) == 1
        assert key not in src_job.document
        assert key in dst_job.data
        assert len(dst_job.data) == 1
        assert key not in src_job.data
        with pytest.raises(RuntimeError):
            src_job.statepoint = dst
        with pytest.raises(DestinationExistsError):
            src_job.statepoint = dst

    @pytest.mark.filterwarnings("ignore:reset_statepoint")
    @pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
    def test_update_statepoint(self):
        key = "move_job"
        d = "test_data"
        src = test_token
        extension = {"dst": True}
        dst = dict(src)
        dst.update(extension)
        extension2 = {"dst": False}
        dst2 = dict(src)
        dst2.update(extension2)
        src_job = self.open_job(src)
        src_job.document[key] = d
        assert key in src_job.document
        assert len(src_job.document) == 1
        src_job.data[key] = d
        assert key in src_job.data
        assert len(src_job.data) == 1
        src_job.update_statepoint(extension)
        src_job = self.open_job(src)
        dst_job = self.open_job(dst)
        assert dst_job.statepoint() == dst
        assert key in dst_job.document
        assert len(dst_job.document) == 1
        assert key not in src_job.document
        assert key in dst_job.data
        assert len(dst_job.data) == 1
        assert key not in src_job.data
        with pytest.raises(RuntimeError):
            src_job.statepoint = dst
        with pytest.raises(DestinationExistsError):
            src_job.statepoint = dst
        with pytest.raises(KeyError):
            dst_job.update_statepoint(extension2)
        dst_job.update_statepoint(extension2, overwrite=True)
        dst2_job = self.open_job(dst2)
        assert dst2_job.statepoint() == dst2
        assert key in dst2_job.document
        assert len(dst2_job.document) == 1
        assert key in dst2_job.data
        assert len(dst2_job.data) == 1


@pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
class TestJobOpenData(TestJobBase):
    @staticmethod
    @contextmanager
    def open_data(job):
        with job.data:
            yield

    def test_get_set(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert not bool(job.data)
            assert len(job.data) == 0
            assert key not in job.data
            job.data[key] = d
            assert bool(job.data)
            assert len(job.data) == 1
            assert key in job.data
            assert job.data[key] == d
            assert job.data.get(key) == d
            assert job.data.get("non-existent-key", d) == d

    def test_del(self):
        key0 = "del0"
        key1 = "del1"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.data) == 0
            assert key0 not in job.data
            job.data[key0] = d0
            assert len(job.data) == 1
            assert key0 in job.data
            job.data[key1] = d1
            assert len(job.data) == 2
            assert key0 in job.data
            assert key1 in job.data
            assert job.data[key0] == d0
            assert job.data[key1] == d1
            del job.data[key0]
            assert len(job.data) == 1
            assert key1 in job.data
            assert key0 not in job.data

    def test_get_set_data(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert not bool(job.data)
            assert len(job.data) == 0
            assert key not in job.data
            job.data[key] = d
            assert bool(job.data)
            assert len(job.data) == 1
            assert key in job.data
            assert job.data[key] == d
            assert job.data.get(key) == d
            assert job.data.get("non-existent-key", d) == d

    def test_set_set_data(self):
        key0 = "set_set0"
        key1 = "set_set1"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert not bool(job.data)
            assert len(job.data) == 0
            assert key0 not in job.data
            job.data[key0] = d0
            assert bool(job.data)
            assert len(job.data) == 1
            assert key0 in job.data
            assert job.data[key0] == d0
        job = self.open_job(test_token)
        with self.open_data(job):
            assert bool(job.data)
            assert len(job.data) == 1
            assert key0 in job.data
            assert job.data[key0] == d0
        job = self.open_job(test_token)
        with self.open_data(job):
            job.data[key1] = d1
            assert bool(job.data)
            assert len(job.data) == 2
            assert key0 in job.data
            assert key1 in job.data
            assert job.data[key0] == d0
            assert job.data[key1] == d1

    def test_get_set_nested(self):
        d0 = "test_data0"
        d1 = "test_data1"
        d2 = "test_data2"
        assert d0 != d1 != d2
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.data) == 0
            assert "key0" not in job.data
            job.data["key0"] = d0
            assert len(job.data) == 1
            assert "key0" in job.data
            assert job.data["key0"] == d0
            with pytest.raises(AttributeError):
                job.data.key0.key1
            job.data.key0 = {"key1": d0}
            assert len(job.data) == 1
            assert "key0" in job.data
            assert dict(job.data) == {"key0": {"key1": d0}}
            assert job.data["key0"] == {"key1": d0}
            assert job.data["key0"]["key1"] == d0
            assert job.data.key0 == {"key1": d0}
            assert job.data.key0.key1 == d0
            job.data.key0.key1 = d1
            assert job.data == {"key0": {"key1": d1}}
            assert job.data["key0"] == {"key1": d1}
            assert job.data["key0"]["key1"] == d1
            assert job.data.key0 == {"key1": d1}
            assert job.data.key0.key1 == d1
            job.data["key0"]["key1"] = d2
            assert job.data == {"key0": {"key1": d2}}
            assert job.data["key0"] == {"key1": d2}
            assert job.data["key0"]["key1"] == d2
            assert job.data.key0 == {"key1": d2}
            assert job.data.key0.key1 == d2

    def test_get_set_nested_data(self):
        d0 = "test_data0"
        d1 = "test_data1"
        d2 = "test_data2"
        assert d0 != d1 != d2
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.data) == 0
            assert "key0" not in job.data
            job.data["key0"] = d0
            assert len(job.data) == 1
            assert "key0" in job.data
            assert job.data["key0"] == d0
            with pytest.raises(AttributeError):
                job.data.key0.key1
            job.data.key0 = {"key1": d0}
            assert len(job.data) == 1
            assert "key0" in job.data
            assert dict(job.data) == {"key0": {"key1": d0}}
            assert job.data["key0"] == {"key1": d0}
            assert job.data["key0"]["key1"] == d0
            assert job.data.key0 == {"key1": d0}
            assert job.data.key0.key1 == d0
            job.data.key0.key1 = d1
            assert job.data == {"key0": {"key1": d1}}
            assert job.data["key0"] == {"key1": d1}
            assert job.data["key0"]["key1"] == d1
            assert job.data.key0 == {"key1": d1}
            assert job.data.key0.key1 == d1
            job.data["key0"]["key1"] = d2
            assert job.data == {"key0": {"key1": d2}}
            assert job.data["key0"] == {"key1": d2}
            assert job.data["key0"]["key1"] == d2
            assert job.data.key0 == {"key1": d2}
            assert job.data.key0.key1 == d2

    def test_assign(self):
        key = "assign"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.data) == 0
            job.data[key] = d0
            assert len(job.data) == 1
            assert dict(job.data) == {key: d0}
            with pytest.raises(ValueError):
                job.data = d1
        job.data = {key: d1}
        assert len(job.data) == 1
        assert dict(job.data) == {key: d1}

    def test_assign_data(self):
        key = "assign"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.data) == 0
            job.data[key] = d0
            assert len(job.data) == 1
            assert dict(job.data) == {key: d0}
            with pytest.raises(ValueError):
                job.data = d1
        job.data = {key: d1}
        assert len(job.data) == 1
        assert dict(job.data) == {key: d1}

    def test_copy_data(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.data[key] = d
            assert bool(job.data)
            assert len(job.data) == 1
            assert key in job.data
            assert job.data[key] == d
            assert job.data.get(key) == d
            assert job.data.get("non-existent-key", d) == d
            copy = dict(job.data)
            assert bool(copy)
            assert len(copy) == 1
            assert key in copy
            assert copy[key] == d
            assert copy.get(key) == d
            assert copy.get("non-existent-key", d) == d

    def test_update(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.data.update({key: d})
            assert key in job.data

    def test_clear_data(self):
        key = "clear"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.data[key] = d
            assert key in job.data
            assert len(job.data) == 1
            job.data.clear()
            assert key not in job.data
            assert len(job.data) == 0

    def test_reopen(self):
        key = "clear"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.data[key] = d
            assert key in job.data
            assert len(job.data) == 1
        job2 = self.open_job(test_token)
        with self.open_data(job2):
            assert key in job2.data
            assert len(job2.data) == 1

    def test_concurrency(self):
        key = "concurrent"
        d = "test_data"
        job = self.open_job(test_token)
        job2 = self.open_job(test_token)
        with self.open_data(job):
            with self.open_data(job2):
                assert key not in job.data
                assert key not in job2.data
                job.data[key] = d
                assert key in job.data
                assert key in job2.data

    def test_move_not_initialized(self):
        job = self.open_job(test_token)
        with pytest.raises(RuntimeError):
            job.move(job._project)

    def test_move_intra_project(self):
        job = self.open_job(test_token).init()
        job.move(self.project)  # no-op

    def test_move_inter_project(self):
        job = self.open_job(test_token).init()
        project_a = self.project
        project_b = self.project_class.init_project(
            path=os.path.join(self._tmp_pr, "project_b")
        )
        job.move(project_b)
        job.move(project_a)
        project_b.clone(job)
        with pytest.raises(DestinationExistsError):
            job.move(project_b)

    def test_remove(self):
        key = "remove"
        job = self.open_job(test_token)
        job.remove()
        d = "test_data"
        with self.open_data(job):
            job.data[key] = d
            assert key in job.data
            assert len(job.data) == 1
        fn_test = os.path.join(job.path, "test")
        with open(fn_test, "w") as file:
            file.write("test")
        assert os.path.isfile(fn_test)
        job.remove()
        with self.open_data(job):
            assert key not in job.data
        assert not os.path.isfile(fn_test)

    def test_clear_job(self):
        key = "clear"
        job = self.open_job(test_token)
        assert job not in self.project
        job.clear()
        assert job not in self.project
        job.clear()
        assert job not in self.project
        job.init()
        assert job in self.project
        job.clear()
        assert job in self.project
        job.clear()
        job.clear()
        assert job in self.project
        d = "test_data"
        with self.open_data(job):
            job.data[key] = d
            assert job in self.project
            assert key in job.data
            assert len(job.data) == 1
        job.clear()
        with self.open_data(job):
            assert len(job.data) == 0
        with open(job.fn("test"), "w") as file:
            file.write("test")
        assert job.isfile("test")
        assert job in self.project
        job.clear()
        assert not job.isfile("test")
        with self.open_data(job):
            assert len(job.data) == 0

    def test_reset(self):
        key = "reset"
        job = self.open_job(test_token)
        assert job not in self.project
        job.reset()
        assert job in self.project
        with self.open_data(job):
            assert len(job.data) == 0
            job.data[key] = "test_data"
            assert len(job.data) == 1
        job.reset()
        assert job in self.project
        with self.open_data(job):
            assert len(job.data) == 0

    def test_data(self):
        key = "test_data"
        job = self.open_job(test_token)

        def check_content(key, d):
            assert job.data[key] == d
            assert getattr(job.data, key) == d
            assert dict(job.data)[key] == d
            assert job.data[key] == d
            assert getattr(job.data, key) == d
            assert dict(job.data)[key] == d

        with self.open_data(job):
            d0 = "test_data0"
            job.data[key] = d0
            check_content(key, d0)
            d1 = "test_data1"
            job.data[key] = d1
            check_content(key, d1)
            d2 = "test_data2"
            job.data[key] = d2
            check_content(key, d2)
            d3 = "test_data3"
            setattr(job.data, key, d3)
            check_content(key, d3)

    def test_reset_statepoint_job(self):
        key = "move_job"
        d = "test_data"
        src = test_token
        dst = dict(test_token)
        dst["dst"] = True
        src_job = self.open_job(src)
        with self.open_data(src_job):
            src_job.data[key] = d
            assert key in src_job.data
            assert len(src_job.data) == 1
        src_job.statepoint = dst
        src_job = self.open_job(src)
        dst_job = self.open_job(dst)
        with self.open_data(dst_job):
            assert key in dst_job.data
            assert len(dst_job.data) == 1
        with self.open_data(src_job):
            assert key not in src_job.data
        with pytest.raises(RuntimeError):
            src_job.statepoint = dst
        with pytest.raises(DestinationExistsError):
            src_job.statepoint = dst

    def test_update_statepoint(self):
        key = "move_job"
        d = "test_data"
        src = test_token
        extension = {"dst": True}
        dst = dict(src)
        dst.update(extension)
        extension2 = {"dst": False}
        dst2 = dict(src)
        dst2.update(extension2)
        src_job = self.open_job(src)
        with self.open_data(src_job):
            src_job.data[key] = d
            assert key in src_job.data
            assert len(src_job.data) == 1
        src_job.update_statepoint(extension)
        src_job = self.open_job(src)
        dst_job = self.open_job(dst)
        assert dst_job.statepoint() == dst
        with self.open_data(dst_job):
            assert key in dst_job.data
            assert len(dst_job.data) == 1
        with self.open_data(src_job):
            assert key not in src_job.data
        with pytest.raises(RuntimeError):
            src_job.statepoint = dst
        with pytest.raises(DestinationExistsError):
            src_job.statepoint = dst
        with pytest.raises(KeyError):
            dst_job.update_statepoint(extension2)
        dst_job.update_statepoint(extension2, overwrite=True)
        dst2_job = self.open_job(dst2)
        assert dst2_job.statepoint() == dst2
        with self.open_data(dst2_job):
            assert key in dst2_job.data
            assert len(dst2_job.data) == 1

    def test_statepoint_copy(self):
        job = self.open_job(dict(a=test_token, b=test_token)).init()
        id_ = job.id
        sp_copy = copy.copy(job.sp)
        del sp_copy["b"]
        assert "a" in job.sp
        assert "b" not in job.sp
        assert job in self.project
        assert job.id != id_

    def test_statepoint_deepcopy(self):
        job = self.open_job(dict(a=test_token, b=test_token)).init()
        id_ = job.id
        sp_copy = copy.deepcopy(job.sp)
        del sp_copy["b"]
        assert "a" in job.sp
        assert "b" in job.sp
        assert job not in self.project
        assert job.id == id_


@pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
class TestJobClosedData(TestJobOpenData):
    @staticmethod
    @contextmanager
    def open_data(job):
        yield

    def test_implicit_initialization(self):
        job = self.open_job(test_token)
        assert "test" not in job.stores
        assert "foo" not in job.stores.test
        assert list(job.stores.keys()) == []
        assert list(job.stores) == []
        assert "test" not in job.stores
        job.stores.test.foo = True
        assert "test" in job.stores
        assert "foo" in job.stores.test
        assert list(job.stores.keys()) == ["test"]
        assert list(job.stores) == ["test"]


@pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
class TestJobOpenCustomData(TestJobBase):
    @staticmethod
    @contextmanager
    def open_data(job):
        with job.stores.test:
            yield

    def test_get_set(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert not bool(job.stores.test)
            assert len(job.stores.test) == 0
            assert key not in job.stores.test
            job.stores.test[key] = d
            assert bool(job.stores.test)
            assert len(job.stores.test) == 1
            assert key in job.stores.test
            assert job.stores.test[key] == d
            assert job.stores.test.get(key) == d
            assert job.stores.test.get("non-existent-key", d) == d

    def test_del(self):
        key0 = "del0"
        key1 = "del1"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.stores.test) == 0
            assert key0 not in job.stores.test
            job.stores.test[key0] = d0
            assert len(job.stores.test) == 1
            assert key0 in job.stores.test
            job.stores.test[key1] = d1
            assert len(job.stores.test) == 2
            assert key0 in job.stores.test
            assert key1 in job.stores.test
            assert job.stores.test[key0] == d0
            assert job.stores.test[key1] == d1
            del job.stores.test[key0]
            assert len(job.stores.test) == 1
            assert key1 in job.stores.test
            assert key0 not in job.stores.test

    def test_get_set_data(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert not bool(job.stores.test)
            assert len(job.stores.test) == 0
            assert key not in job.stores.test
            job.stores.test[key] = d
            assert bool(job.stores.test)
            assert len(job.stores.test) == 1
            assert key in job.stores.test
            assert job.stores.test[key] == d
            assert job.stores.test.get(key) == d
            assert job.stores.test.get("non-existent-key", d) == d

    def test_set_set_data(self):
        key0 = "set_set0"
        key1 = "set_set1"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert not bool(job.stores.test)
            assert len(job.stores.test) == 0
            assert key0 not in job.stores.test
            job.stores.test[key0] = d0
            assert bool(job.stores.test)
            assert len(job.stores.test) == 1
            assert key0 in job.stores.test
            assert job.stores.test[key0] == d0
        job = self.open_job(test_token)
        with self.open_data(job):
            assert bool(job.stores.test)
            assert len(job.stores.test) == 1
            assert key0 in job.stores.test
            assert job.stores.test[key0] == d0
        job = self.open_job(test_token)
        with self.open_data(job):
            job.stores.test[key1] = d1
            assert bool(job.stores.test)
            assert len(job.stores.test) == 2
            assert key0 in job.stores.test
            assert key1 in job.stores.test
            assert job.stores.test[key0] == d0
            assert job.stores.test[key1] == d1

    def test_get_set_nested(self):
        d0 = "test_data0"
        d1 = "test_data1"
        d2 = "test_data2"
        assert d0 != d1 != d2
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.stores.test) == 0
            assert "key0" not in job.stores.test
            job.stores.test["key0"] = d0
            assert len(job.stores.test) == 1
            assert "key0" in job.stores.test
            assert job.stores.test["key0"] == d0
            with pytest.raises(AttributeError):
                job.stores.test.key0.key1
            job.stores.test.key0 = {"key1": d0}
            assert len(job.stores.test) == 1
            assert "key0" in job.stores.test
            assert dict(job.stores.test) == {"key0": {"key1": d0}}
            assert job.stores.test["key0"] == {"key1": d0}
            assert job.stores.test["key0"]["key1"] == d0
            assert job.stores.test.key0 == {"key1": d0}
            assert job.stores.test.key0.key1 == d0
            job.stores.test.key0.key1 = d1
            assert job.stores.test == {"key0": {"key1": d1}}
            assert job.stores.test["key0"] == {"key1": d1}
            assert job.stores.test["key0"]["key1"] == d1
            assert job.stores.test.key0 == {"key1": d1}
            assert job.stores.test.key0.key1 == d1
            job.stores.test["key0"]["key1"] = d2
            assert job.stores.test == {"key0": {"key1": d2}}
            assert job.stores.test["key0"] == {"key1": d2}
            assert job.stores.test["key0"]["key1"] == d2
            assert job.stores.test.key0 == {"key1": d2}
            assert job.stores.test.key0.key1 == d2

    def test_get_set_nested_data(self):
        d0 = "test_data0"
        d1 = "test_data1"
        d2 = "test_data2"
        assert d0 != d1 != d2
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.stores.test) == 0
            assert "key0" not in job.stores.test
            job.stores.test["key0"] = d0
            assert len(job.stores.test) == 1
            assert "key0" in job.stores.test
            assert job.stores.test["key0"] == d0
            with pytest.raises(AttributeError):
                job.stores.test.key0.key1
            job.stores.test.key0 = {"key1": d0}
            assert len(job.stores.test) == 1
            assert "key0" in job.stores.test
            assert dict(job.stores.test) == {"key0": {"key1": d0}}
            assert job.stores.test["key0"] == {"key1": d0}
            assert job.stores.test["key0"]["key1"] == d0
            assert job.stores.test.key0 == {"key1": d0}
            assert job.stores.test.key0.key1 == d0
            job.stores.test.key0.key1 = d1
            assert job.stores.test == {"key0": {"key1": d1}}
            assert job.stores.test["key0"] == {"key1": d1}
            assert job.stores.test["key0"]["key1"] == d1
            assert job.stores.test.key0 == {"key1": d1}
            assert job.stores.test.key0.key1 == d1
            job.stores.test["key0"]["key1"] = d2
            assert job.stores.test == {"key0": {"key1": d2}}
            assert job.stores.test["key0"] == {"key1": d2}
            assert job.stores.test["key0"]["key1"] == d2
            assert job.stores.test.key0 == {"key1": d2}
            assert job.stores.test.key0.key1 == d2

    def test_assign(self):
        key = "assign"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.stores.test) == 0
            job.stores.test[key] = d0
            assert len(job.stores.test) == 1
            assert dict(job.stores.test) == {key: d0}
            with pytest.raises(ValueError):
                job.stores.test = d1
        job.stores.test = {key: d1}
        assert len(job.stores.test) == 1
        assert dict(job.stores.test) == {key: d1}

    def test_assign_data(self):
        key = "assign"
        d0 = "test_data0"
        d1 = "test_data1"
        job = self.open_job(test_token)
        with self.open_data(job):
            assert len(job.stores.test) == 0
            job.stores.test[key] = d0
            assert len(job.stores.test) == 1
            assert dict(job.stores.test) == {key: d0}
            with pytest.raises(ValueError):
                job.stores.test = d1
        job.stores.test = {key: d1}
        assert len(job.stores.test) == 1
        assert dict(job.stores.test) == {key: d1}

    def test_copy_data(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.stores.test[key] = d
            assert bool(job.stores.test)
            assert len(job.stores.test) == 1
            assert key in job.stores.test
            assert job.stores.test[key] == d
            assert job.stores.test.get(key) == d
            assert job.stores.test.get("non-existent-key", d) == d
            copy = dict(job.stores.test)
            assert bool(copy)
            assert len(copy) == 1
            assert key in copy
            assert copy[key] == d
            assert copy.get(key) == d
            assert copy.get("non-existent-key", d) == d

    def test_update(self):
        key = "get_set"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.stores.test.update({key: d})
            assert key in job.stores.test

    def test_clear_data(self):
        key = "clear"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.stores.test[key] = d
            assert key in job.stores.test
            assert len(job.stores.test) == 1
            job.stores.test.clear()
            assert key not in job.stores.test
            assert len(job.stores.test) == 0

    def test_reopen(self):
        key = "reopen"
        d = "test_data"
        job = self.open_job(test_token)
        with self.open_data(job):
            job.stores.test[key] = d
            assert key in job.stores.test
            assert len(job.stores.test) == 1
        job2 = self.open_job(test_token)
        with self.open_data(job2):
            assert key in job2.stores.test
            assert len(job2.stores.test) == 1

    def test_concurrency(self):
        key = "concurrent"
        d = "test_data"
        job = self.open_job(test_token)
        job2 = self.open_job(test_token)
        with self.open_data(job):
            with self.open_data(job2):
                assert key not in job.stores.test
                assert key not in job2.stores.test
                job.stores.test[key] = d
                assert key in job.stores.test
                assert key in job2.stores.test

    def test_remove(self):
        key = "remove"
        job = self.open_job(test_token)
        job.remove()
        d = "test_data"
        with self.open_data(job):
            job.stores.test[key] = d
            assert key in job.stores.test
            assert len(job.stores.test) == 1
        fn_test = os.path.join(job.path, "test")
        with open(fn_test, "w") as file:
            file.write("test")
        assert os.path.isfile(fn_test)
        job.remove()
        with self.open_data(job):
            assert key not in job.stores.test
        assert not os.path.isfile(fn_test)

    def test_clear_job(self):
        key = "clear"
        job = self.open_job(test_token)
        assert job not in self.project
        job.clear()
        assert job not in self.project
        job.clear()
        assert job not in self.project
        job.init()
        assert job in self.project
        job.clear()
        assert job in self.project
        job.clear()
        job.clear()
        assert job in self.project
        d = "test_data"
        with self.open_data(job):
            job.stores.test[key] = d
            assert job in self.project
            assert key in job.stores.test
            assert len(job.stores.test) == 1
        job.clear()
        with self.open_data(job):
            assert len(job.stores.test) == 0
        with open(job.fn("test"), "w") as file:
            file.write("test")
        assert job.isfile("test")
        assert job in self.project
        job.clear()
        assert not job.isfile("test")
        with self.open_data(job):
            assert len(job.stores.test) == 0

    def test_reset(self):
        key = "reset"
        job = self.open_job(test_token)
        assert job not in self.project
        job.reset()
        assert job in self.project
        with self.open_data(job):
            assert len(job.stores.test) == 0
            job.stores.test[key] = "test_data"
            assert len(job.stores.test) == 1
        job.reset()
        assert job in self.project
        with self.open_data(job):
            assert len(job.stores.test) == 0

    def test_data(self):
        key = "test_data"
        job = self.open_job(test_token)

        def check_content(key, d):
            assert job.stores.test[key] == d
            assert getattr(job.stores.test, key) == d
            assert dict(job.stores.test)[key] == d
            assert job.stores.test[key] == d
            assert getattr(job.stores.test, key) == d
            assert dict(job.stores.test)[key] == d

        with self.open_data(job):
            d0 = "test_data0"
            job.stores.test[key] = d0
            check_content(key, d0)
            d1 = "test_data1"
            job.stores.test[key] = d1
            check_content(key, d1)
            d2 = "test_data2"
            job.stores.test[key] = d2
            check_content(key, d2)
            d3 = "test_data3"
            setattr(job.stores.test, key, d3)
            check_content(key, d3)


class TestJobClosedCustomData(TestJobOpenCustomData):
    @staticmethod
    @contextmanager
    def open_data(job):
        yield
